"""Dynamic Convolution modules."""

import typing as T

import torch
from torch import nn

from unipercept.log import logger
from unipercept.nn._args import to_2tuple
from unipercept.nn.activations import ActivationSpec, InplaceReLU
from unipercept.nn.init import init_xavier_fill_
from unipercept.nn.linear import MLP
from unipercept.nn.norms import NormSpec, get_norm
from unipercept.ops.dynconv import dynconv2d
from unipercept.types import Tensor

########################
# Modules and wrappers #
########################


class DynamicConv2d(nn.Module):
    """
    Module that performs a dynamic convolution between a set of kernels and a feature map.

    The kernels are a set of K vectors, which are mapped and reshaped to K tensors that
    are convolved with the feature map of shape (... H W).

    The output is a tensor of shape (... K H W).
    """

    def __init__(self) -> None:
        super().__init__()

    @T.override
    def forward(self, features: Tensor, kernels: Tensor) -> Tensor:
        """Perform dynamic convolution on the features using the flattened kernels."""
        return dynconv2d(features, kernels)


class ProjectConv2d(nn.Module):
    """Dynamic convolution over 2D feature map, with optional projection layers.

    Parameters
    ----------
    d_f : int
        The number of dimensions in the feature tensor.
    d_k : int
        The number of dimensions in the kernel tensor.
    d_h : int
        The number of hidden dimensions.
    layers : int | Tuple[int,int] | Sequence[int]
        The number of layers in the projection MLPs for the feature and kernel,
        respectively. If a single integer is provided, the same number of layers
        is used for both projections.
    layer_expansion : float | Tuple[float,float] | Sequence[float] | None
        The expansion factor to use in the projection layers. If a single float is
        provided, the same factor is used for both projections. When no layers
        are used, the expansion factor must be explicitly set to None.
    activation: ActivationSpec
        The activation function to use in projection layers.
    dropout : float
        The dropout rate in the projection layers.
    input_norm : NormSpec | Tuple[NormSpec,NormSpec] | Sequence[None]
        The normalization layer to apply to the input tensors. If a single norm
        is provided, the same norm is used for both inputs. If None, no normalization
        is applied.
    init_gain : float | Tuple[float,float] | Sequence[float]
        The gain to apply to the output tensor initialization. When no layers
        are used, the expansion factor must be explicitly set to None.
    """

    feature_norm: nn.Module | None
    feature_ffn: nn.Module | None

    kernel_norm: nn.Module | None
    kernel_ffn: nn.Module | None

    def __init__(
        self,
        d_feature: int,
        d_kernel: int,
        d_hidden: int,
        layers: int | tuple[int, int] | T.Iterable[int] = (1, 3),
        layer_expansion: (
            float | tuple[float | None, float | None] | T.Sequence[float | None] | None
        ) = 2.0,
        activation: ActivationSpec = InplaceReLU,
        dropout: float = 0.0,
        input_norm: (
            NormSpec
            | tuple[NormSpec | None, NormSpec | None]
            | T.Sequence[NormSpec | None]
            | None
        ) = None,
        init_gain: (
            float | tuple[float | None, float | None] | T.Sequence[float | None] | None
        ) = (0.33, 0.33),
        bias: tuple[bool | None, bool | None] | bool | None = None,
    ) -> None:
        super().__init__()

        self.d_f = d_feature
        self.d_k = d_kernel
        self.d_o = d_hidden

        # Generate layers for each branch
        for (
            br_name,
            br_dim,
            br_layers,
            br_gain,
            br_expansion,
            br_dropout,
            br_norm,
            br_activation,
            br_bias,
        ) in zip(
            ("feature", "kernel"),
            (d_feature, d_kernel),
            to_2tuple(layers),
            to_2tuple(init_gain),
            to_2tuple(layer_expansion),
            to_2tuple(dropout),
            to_2tuple(input_norm),
            to_2tuple(activation),
            to_2tuple(bias),
            strict=True,
        ):
            norm = ffn = None

            if br_norm is not None:
                norm = get_norm(br_norm, br_dim)
            if br_layers > 1:
                ffn = MLP(
                    br_dim,
                    d_hidden,
                    hidden_features=br_expansion,
                    bias=None,
                    layers=br_layers,
                    activation=br_activation,
                    dropout=br_dropout,
                    init_gain=br_gain,
                )
            else:
                if br_layers == 1:
                    ffn = nn.Linear(br_dim, d_hidden, bias=bool(br_bias))
                    init_xavier_fill_(ffn)
                    if br_gain is not None:
                        ffn.weight.data.normal_(mean=0.0, std=br_gain)
                elif br_dim != d_hidden:
                    msg = (
                        f"{br_name} layers must be non-negative to map input dimension "
                        f"({br_dim}) to hidden dimension ({d_hidden})."
                    )
                    raise ValueError(msg)
                elif br_gain is not None and br_gain != 1.0:
                    logger.warning(f"No layers used, {br_name} init gain ignored.")

                # Warn the user about unused arguments (should be None to avoid this)
                if br_expansion is not None and br_expansion != 1.0:
                    logger.warning(
                        f"Less than two layers used, {br_name} expansion ignored."
                    )
                if br_dropout > 0.0:
                    logger.warning(
                        f"Less than two layers used, {br_name} dropout ignored."
                    )
                if br_activation is not None:
                    logger.warning(
                        f"Less than two layers used, {br_name} activation ignored."
                    )

            self.register_module(f"{br_name}_norm", norm)
            self.register_module(f"{br_name}_ffn", ffn)

    @T.override
    def forward(self, f: Tensor, k: Tensor) -> Tensor:
        r"""
        Perform dynamic convolution on the feature tensor.

        Parameters
        ----------
        f : Tensor[N, C_f, H, W]
            The feature tensor to convolve.
        k : Tensor[N, K, C_k]
            The kernel tensor to convolve with.

        Returns
        -------
        Tensor[N, K, H, W]
            The convolved tensor.
        """
        _, _, h, w = f.shape

        # Feature projection
        # f = rearrange(f, "... c h w -> ... (h w) c", h=h, w=w)
        f = f.flatten(2)  # (B, C, H*W)
        f = f.transpose(1, 2)  # (B, H*W, C)

        if self.feature_norm is not None:
            f = self.feature_norm(f)
        if self.feature_ffn is not None:
            f = self.feature_ffn(f)

        # Kernel projection
        if self.kernel_norm is not None:
            k = self.kernel_norm(k)
        if self.kernel_ffn is not None:
            k = self.kernel_ffn(k)

        # Dynamic convolution
        # o = einsum(k, f, "... n c, ... hw c -> ... n hw")
        o = torch.einsum("bkc,bsc->bks", k, f)
        # o = rearrange(o, "... n (h w) -> ... n h w", h=h, w=w)
        o = o.unflatten(-1, (h, w))
        return o.contiguous()



from __future__ import annotations

from enum import Enum
from typing import Any

import torch
from torch import Tensor, device

from ._se3 import se3_log_map


@torch.jit.script_if_tracing
def _safe_det_3x3(t: Tensor):
    """
    Fast determinant calculation for a batch of 3x3 matrices.

    Note, result of this function might not be the same as `torch.det()`.
    The differences might be in the last significant digit.

    Args:
        t: Tensor of shape (N, 3, 3).

    Returns
    -------
        Tensor of shape (N) with determinants.
    """
    return (
        t[..., 0, 0] * (t[..., 1, 1] * t[..., 2, 2] - t[..., 1, 2] * t[..., 2, 1])
        - t[..., 0, 1] * (t[..., 1, 0] * t[..., 2, 2] - t[..., 2, 0] * t[..., 1, 2])
        + t[..., 0, 2] * (t[..., 1, 0] * t[..., 2, 1] - t[..., 2, 0] * t[..., 1, 1])
    )


@torch.jit.script_if_tracing
def _broadcast_bmm(a: Tensor, b: Tensor) -> Tensor:
    """
    Batch multiply two matrices and broadcast if necessary.

    Args:
        a: torch tensor of shape (P, K) or (M, P, K)
        b: torch tensor of shape (N, K, K)

    Returns
    -------
        a and b broadcast multiplied. The output batch dimension is max(N, M).

    To broadcast transforms across a batch dimension if M != N then
    expect that either M = 1 or N = 1. The tensor with batch dimension 1 is
    expanded to have shape N or M.
    """
    if a.dim() == 2:
        a = a[None]
    if len(a) != len(b):
        if not ((len(a) == 1) or (len(b) == 1)):
            msg = "Expected batch dim for bmm to be equal or 1; got %r, %r"
            raise ValueError(msg % (a.shape, b.shape))
        if len(a) == 1:
            a = a.expand(len(b), -1, -1)
        if len(b) == 1:
            b = b.expand(len(a), -1, -1)
    return a.bmm(b)


@torch.jit.script_if_tracing
def _axis_angle_rotation(axis: str, angle: Tensor) -> Tensor:
    """
    Return the rotation matrices for one of the rotations about an axis
    of which Euler angles describe, for each value of the angle given.

    Args:
        axis: Axis label "X" or "Y or "Z".
        angle: any shape tensor of Euler angles in radians

    Returns
    -------
        Rotation matrices as tensor of shape (..., 3, 3).
    """
    cos = torch.cos(angle)
    sin = torch.sin(angle)
    one = torch.ones_like(angle)
    zero = torch.zeros_like(angle)

    if axis == "X":
        R_flat = (one, zero, zero, zero, cos, -sin, zero, sin, cos)
    elif axis == "Y":
        R_flat = (cos, zero, sin, zero, one, zero, -sin, zero, cos)
    elif axis == "Z":
        R_flat = (cos, -sin, zero, sin, cos, zero, zero, zero, one)
    else:
        msg = "letter must be either X, Y or Z."
        raise ValueError(msg)

    return torch.stack(R_flat, -1).reshape((*angle.shape, 3, 3))


class TransformType(Enum):
    """
    Our implementations depends on TorchScript, for this reason we cannot support
    subclassing of the Transform3D into Rotate, Scale, etc.
    This enumeration is used instead to hardcode the inverse method.
    """

    DEFAULT = 0
    TRANSLATE = 1
    SCALE = 2
    ROTATE = 3


@torch.jit.script_if_tracing
def _inverse(M: Tensor, tt: TransformType) -> Tensor:
    if tt == TransformType.TRANSLATE:
        inv_mask = M.new_ones([1, 4, 4])
        inv_mask[0, 3, :3] = -1.0
        return M * inv_mask
    if tt == TransformType.SCALE:
        xyz = torch.stack([M[:, i, i] for i in range(4)], dim=1)
        ixyz = 1.0 / xyz
        return torch.diag_embed(ixyz, dim1=1, dim2=2)
    if tt == TransformType.ROTATE:
        return M.permute(0, 2, 1).contiguous()
    return torch.inverse(M)


@torch.jit.script
class Transform3d:
    def __init__(self, M: Tensor, *, tt: TransformType) -> None:
        """
        Parameters
        ----------
        matrix
            A tensor of shape (4, 4) or of shape (minibatch, 4, 4) representing
            the 4x4 3D transformation matrix.
        """
        if M.ndim not in (2, 3):
            msg = f"'matrix' has to be a 2- or a 3-dimensional tensor, got {M.ndim}!"
            raise ValueError(msg)
        if M.shape[-2] != 4 or M.shape[-1] != 4:
            msg = (
                f"'matrix' has to be a tensor of shape (minibatch, 4, 4) or "
                f"(4, 4), got {M.shape}!"
            )
            raise ValueError(msg)

        self._matrix = M.view(-1, 4, 4)
        self._transforms: list[Any] = []
        self._tt = tt

    def identity_like(self) -> Transform3d:
        """
        Returns a transform with an identity matrix that has the same datatype
        and device as this transform.
        """
        dtype = self._matrix.dtype
        device = self._matrix.device
        return Transform3d(
            torch.eye(4, dtype=dtype, device=device).view(1, 4, 4),
            tt=TransformType.DEFAULT,
        )

    def __len__(self) -> int:
        return self.get_matrix().shape[0]

    def compose(self, others: list[Transform3d]) -> Transform3d:
        """
        Return a new Transform3d representing the composition of self with the
        given other transforms, which will be stored as an internal list.

        Parameters
        ----------
        *others
            Any number of Transform3d objects

        Returns
        -------
            A new Transform3d with the stored transforms
        """
        out = self.clone()
        for t in others:
            assert isinstance(t, Transform3d)
            out._transforms.append(t)

        return out

    def get_matrix(self) -> Tensor:
        """
        Returns a 4×4 matrix corresponding to each transform in the batch.

        If the transform was composed from others, the matrix for the composite
        transform will be returned.
        For example, if self.transforms contains transforms t1, t2, and t3, and
        given a set of points x, the following should be true:

        .. code-block:: python

            y1 = t1.compose(t2, t3).transform(x)
            y2 = t3.transform(t2.transform(t1.transform(x)))
            y1.get_matrix() == y2.get_matrix()

        Where necessary, those transforms are broadcast against each other.

        Returns
        -------
            A (N, 4, 4) batch of transformation matrices representing the
            stored transforms. See the class documentation for the conventions.
        """
        composed_matrix: Tensor = self._matrix.clone()
        if len(self._transforms) > 0:
            for other in self._transforms:
                assert isinstance(other, Transform3d)

                # Recursion is not allowed in TorchScript, so we hardcode
                # a maximum depth.
                # other_matrix = other.get_matrix()
                if len(other._transforms) > 0:
                    msg = "Max. transform depth reached!"
                    raise NotImplementedError(msg)

                other_matrix: Tensor = other._matrix.clone()
                # composed_matrix = _broadcast_bmm(composed_matrix, other_matrix)
                composed_matrix = composed_matrix.bmm(other_matrix)
        return composed_matrix

    def get_se3_log(self, eps: float = 1e-4, cos_bound: float = 1e-4) -> Tensor:
        """
        Returns a 6D SE(3) log vector corresponding to each transform in the batch.

        In the SE(3) logarithmic representation SE(3) matrices are
        represented as 6-dimensional vectors `[log_translation | log_rotation]`,
        i.e. a concatenation of two 3D vectors `log_translation` and `log_rotation`.

        The conversion from the 4x4 SE(3) matrix `transform` to the
        6D representation `log_transform = [log_translation | log_rotation]`
        is done as follows:
            ```
            log_transform = log(transform.get_matrix())
            log_translation = log_transform[3, :3]
            log_rotation = inv_hat(log_transform[:3, :3])
            ```
        where `log` is the matrix logarithm
        and `inv_hat` is the inverse of the Hat operator [2].

        See the docstring for `se3.se3_log_map` and [1], Sec 9.4.2. for more
        detailed description.

        Parameters
        ----------
        eps
            A threshold for clipping the squared norm of the rotation logarithm
            to avoid division by zero in the singular case.
        cos_bound
            Clamps the cosine of the rotation angle to
            [-1 + cos_bound, 3 - cos_bound] to avoid non-finite outputs.
            The non-finite outputs can be caused by passing small rotation angles
            to the `acos` function in `so3_rotation_angle` of `so3_log_map`.

        Returns
        -------
            A (N, 6) tensor, rows of which represent the individual transforms
            stored in the object as SE(3) logarithms.

        Raises
        ------
            ValueError if the stored transform is not Euclidean (e.g. R is not a rotation
                matrix or the last column has non-zeros in the first three places).

        [1] https://jinyongjeong.github.io/Download/SE3/jlblanco2010geometry3d_techrep.pdf
        [2] https://en.wikipedia.org/wiki/Hat_operator
        """
        return se3_log_map(self.get_matrix(), eps, cos_bound)

    def _get_matrix_inverse(self) -> torch.Tensor:
        return _inverse(self._matrix, self._tt)

    def inverse(self, invert_composed: bool) -> Transform3d:
        """
        Returns a new Transform3d object that represents an inverse of the
        current transformation.

        Parameters
        ----------
        invert_composed
            - True: First compose the list of stored transformations
                and then apply inverse to the result. This is
                potentially slower for classes of transformations
                with inverses that can be computed efficiently
                (e.g. rotations and translations).
            - False: Invert the individual stored transformations
                independently without composing them.

        Returns
        -------
            A new Transform3d object containing the inverse of the original
            transformation.
        """
        tinv = self.identity_like()

        if invert_composed:
            # first compose then invert
            tinv._matrix = torch.inverse(self.get_matrix())
        else:
            # self._get_matrix_inverse() implements efficient inverse
            # of self._matrix
            i_matrix = self._get_matrix_inverse()

            # 2 cases:
            if len(self._transforms) > 0:
                # a) Either we have a non-empty list of transforms:
                # Here we take self._matrix and append its inverse at the
                # end of the reverted _transforms list. After composing
                # the transformations with get_matrix(), this correctly
                # right-multiplies by the inverse of self._matrix
                # at the end of the composition.
                for t_i in range(len(self._transforms) - 1, -1, -1):
                    t = self._transforms[t_i]
                    assert isinstance(t, Transform3d)

                    # TorchScript does now allow recursion
                    # tinv._transforms.append(t.inverse())
                    assert len(t._transforms) == 0

                    tinv._transforms.append(t._get_matrix_inverse())

                last = self.identity_like()
                last._matrix = i_matrix

                tinv._transforms.append(last)
            else:
                # b) Or there are no stored transformations
                # we just set inverted matrix
                tinv._matrix = i_matrix

        return tinv

    # def stack(self, others: List[Transform3d]) -> Transform3d:
    #     """
    #     Return a new batched Transform3d representing the batch elements from
    #     self and all the given other transforms all batched together.

    #     Args:
    #         *others: Any number of Transform3d objects

    #     Returns:
    #         A new Transform3d.
    #     """
    #     transforms = [self] + others
    #     matrix = torch.cat([t.get_matrix() for t in transforms], dim=0)

    #     return Transform3d(matrix, tt=TransformType.DEFAULT)

    def transform_points(
        self, points: torch.Tensor, eps: float | None = None
    ) -> Tensor:
        """
        Use this transform to transform a set of 3D points. Assumes row major
        ordering of the input points.

        Args:
            points: Tensor of shape (P, 3) or (N, P, 3)
            eps: If eps!=None, the argument is used to clamp the
                last coordinate before performing the final division.
                The clamping corresponds to:
                last_coord := (last_coord.sign() + (last_coord==0)) *
                torch.clamp(last_coord.abs(), eps),
                i.e. the last coordinates that are exactly 0 will
                be clamped to +eps.

        Returns
        -------
            points_out: points of shape (N, P, 3) or (P, 3) depending
            on the dimensions of the transform
        """
        points_batch = points.clone()
        if points_batch.dim() == 2:
            points_batch = points_batch[None]  # (P, 3) -> (1, P, 3)
        if points_batch.dim() != 3:
            msg = "Expected points to have dim = 2 or dim = 3: got shape %r"
            raise ValueError(msg % points.shape)

        N, P, _3 = points_batch.shape
        ones = torch.ones(N, P, 1, dtype=points.dtype, device=points.device)
        points_batch = torch.cat([points_batch, ones], dim=2)

        assert isinstance(self, Transform3d)
        composed_matrix = self.get_matrix()

        points_out = _broadcast_bmm(points_batch, composed_matrix)
        denom = points_out[..., 3:]  # denominator
        if eps is not None:
            denom_sign = denom.sign() + (denom == 0.0).type_as(denom)
            denom = denom_sign * torch.clamp(denom.abs(), eps)
        points_out = points_out[..., :3] / denom

        # When transform is (1, 4, 4) and points is (P, 3) return
        # points_out of shape (P, 3)
        if points_out.shape[0] == 1 and points.dim() == 2:
            points_out = points_out.reshape(points.shape)

        return points_out

    def transform_normals(self, normals) -> Tensor:
        """
        Use this transform to transform a set of normal vectors.

        Args:
            normals: Tensor of shape (P, 3) or (N, P, 3)

        Returns
        -------
            normals_out: Tensor of shape (P, 3) or (N, P, 3) depending
            on the dimensions of the transform
        """
        if normals.dim() not in [2, 3]:
            msg = "Expected normals to have dim = 2 or dim = 3: got shape %r"
            raise ValueError(msg % (normals.shape,))
        composed_matrix = self.get_matrix()

        # TODO: inverse is bad! Solve a linear system instead
        mat = composed_matrix[:, :3, :3]
        normals_out = _broadcast_bmm(normals, torch.inverse(mat.transpose(1, 2)))

        # When transform is (1, 4, 4) and normals is (P, 3) return
        # normals_out of shape (P, 3)
        if normals_out.shape[0] == 1 and normals.dim() == 2:
            normals_out = normals_out.reshape(normals.shape)

        return normals_out

    def clone(self) -> Transform3d:
        """
        Deep copy of Transforms object. All internal tensors are cloned
        individually.

        Returns
        -------
            new Transforms object.
        """
        cloned = Transform3d(self._matrix.clone(), tt=self._tt)
        cloned._matrix = self._matrix.clone()
        for t in self._transforms:
            assert isinstance(t, Transform3d)

            # TorchScript recusion not allowed
            # Unroll manually instead
            # other._transforms.append(t.clone())
            assert len(t._transforms) == 0, "Max. transform depth reached!"

            cloned._transforms.append(Transform3d(t._matrix.clone(), tt=t._tt))

        return cloned


@torch.jit.script_if_tracing
def Translate(
    T,
) -> Transform3d:
    """
    Create a new Transform3d representing 3D translations.

    Option I: Translate(xyz, dtype=torch.float32, device='cpu')
        xyz should be a tensor of shape (N, 3)

    Option II: Translate(x, y, z, dtype=torch.float32, device='cpu')
        Here x, y, and z will be broadcast against each other and
        concatenated to form the translation. Each can be:
            - A python scalar
            - A torch scalar
            - A 1D torch tensor
    """
    N = T.shape[0]

    mat = torch.eye(4, dtype=T.dtype, device=T.device)
    mat = mat.view(1, 4, 4).repeat(N, 1, 1)
    mat[:, 3, :3] = T

    return Transform3d(mat, tt=TransformType.TRANSLATE)


@torch.jit.unused
def Scale(
    x,
    y=None,
    z=None,
    dtype: torch.dtype = torch.float32,
) -> Transform3d:
    """
    A Transform3d representing a scaling operation, with different scale
    factors along each coordinate axis.

    Option I: Scale(s, dtype=torch.float32, device='cpu')
        s can be one of
            - Python scalar or torch scalar: Single uniform scale
            - 1D torch tensor of shape (N,): A batch of uniform scale
            - 2D torch tensor of shape (N, 3): Scale differently along each axis

    Option II: Scale(x, y, z, dtype=torch.float32, device='cpu')
        Each of x, y, and z can be one of
            - python scalar
            - torch scalar
            - 1D torch tensor
    """
    xyz = _handle_input(x, y, z, dtype, device, "scale", allow_singleton=True)
    N = xyz.shape[0]

    # TODO: Can we do this all in one go somehow?
    mat = torch.eye(4, dtype=xyz.dtype, device=xyz.device)
    mat = mat.view(1, 4, 4).repeat(N, 1, 1)
    mat[:, 0, 0] = xyz[:, 0]
    mat[:, 1, 1] = xyz[:, 1]
    mat[:, 2, 2] = xyz[:, 2]
    return Transform3d(mat, tt=TransformType.SCALE)


@torch.jit.script_if_tracing
def Rotate(
    R: Tensor,
) -> Transform3d:
    """
    Create a new Transform3d representing 3D rotation using a rotation
    matrix as the input.

    Args:
        R: a tensor of shape (3, 3) or (N, 3, 3)
        orthogonal_tol: tolerance for the test of the orthogonality of R

    """
    if R.dim() == 2:
        R = R[None]
    if R.shape[-2:] != (3, 3):
        msg = f"R must have shape (3, 3) or (N, 3, 3); got {R.shape}"
        raise ValueError(msg)

    N = R.shape[0]
    mat = torch.eye(4, dtype=R.dtype, device=R.device)
    mat = mat.view(1, 4, 4).repeat(N, 1, 1)
    mat[:, :3, :3] = R
    return Transform3d(mat, tt=TransformType.ROTATE)


@torch.jit.unused
def RotateAxisAngle(
    angle,
    axis: str = "X",
    degrees: bool = True,
    dtype: torch.dtype = torch.float32,
    device: Device | None = None,
) -> Transform3d:
    """
    Create a new Transform3d representing 3D rotation about an axis
    by an angle.

    Assuming a right-hand coordinate system, positive rotation angles result
    in a counter clockwise rotation.

    Args:
        angle:
            - A torch tensor of shape (N,)
            - A python scalar
            - A torch scalar
        axis:
            string: one of ["X", "Y", "Z"] indicating the axis about which
            to rotate.
            NOTE: All batch elements are rotated about the same axis.
    """
    axis = axis.upper()
    if axis not in ["X", "Y", "Z"]:
        msg = "Expected axis to be one of ['X', 'Y', 'Z']; got %s"
        raise ValueError(msg % axis)
    angle = _handle_angle_input(angle, dtype, device, "RotateAxisAngle")
    angle = (angle / 180.0 * torch.pi) if degrees else angle
    # We assume the points on which this transformation will be applied
    # are row vectors. The rotation matrix returned from _axis_angle_rotation
    # is for transforming column vectors. Therefore we transpose this matrix.
    # R will always be of shape (N, 3, 3)
    R = _axis_angle_rotation(axis, angle).transpose(1, 2)

    return Rotate(R)


def _handle_coord(c, dtype: torch.dtype, device: torch.device) -> Tensor:
    """
    Helper function for _handle_input.

    Args:
        c: Python scalar, torch scalar, or 1D torch tensor

    Returns
    -------
        c_vec: 1D torch tensor
    """
    if not torch.is_tensor(c):
        c = torch.tensor(c, dtype=dtype, device=device)
    if c.dim() == 0:
        c = c.view(1)
    if c.device != device or c.dtype != dtype:
        c = c.to(device=device, dtype=dtype)
    return c


def _handle_input(
    x,
    y,
    z,
    dtype: torch.dtype,
    device: Device | None,
    name: str,
    allow_singleton: bool = False,
) -> Tensor:
    """
    Helper function to handle parsing logic for building transforms. The output
    is always a tensor of shape (N, 3), but there are several types of allowed
    input.

    Case I: Single Matrix
        In this case x is a tensor of shape (N, 3), and y and z are None. Here just
        return x.

    Case II: Vectors and Scalars
        In this case each of x, y, and z can be one of the following
            - Python scalar
            - Torch scalar
            - Torch tensor of shape (N, 1) or (1, 1)
        In this case x, y and z are broadcast to tensors of shape (N, 1)
        and concatenated to a tensor of shape (N, 3)

    Case III: Singleton (only if allow_singleton=True)
        In this case y and z are None, and x can be one of the following:
            - Python scalar
            - Torch scalar
            - Torch tensor of shape (N, 1) or (1, 1)
        Here x will be duplicated 3 times, and we return a tensor of shape (N, 3)

    Returns
    -------
        xyz: Tensor of shape (N, 3)
    """
    device_ = get_device(x, device)
    # If x is actually a tensor of shape (N, 3) then just return it
    if torch.is_tensor(x) and x.dim() == 2:
        if x.shape[1] != 3:
            msg = "Expected tensor of shape (N, 3); got %r (in %s)"
            raise ValueError(msg % (x.shape, name))
        if y is not None or z is not None:
            msg = f"Expected y and z to be None (in {name})"
            raise ValueError(msg)
        return x.to(device=device_, dtype=dtype)

    if allow_singleton and y is None and z is None:
        y = x
        z = x

    # Convert all to 1D tensors
    xyz = [_handle_coord(c, dtype, device_) for c in [x, y, z]]

    # Broadcast and concatenate
    sizes = [c.shape[0] for c in xyz]
    N = max(sizes)
    for c in xyz:
        if c.shape[0] != 1 and c.shape[0] != N:
            msg = f"Got non-broadcastable sizes {sizes!r} (in {name})"
            raise ValueError(msg)
    xyz = [c.expand(N) for c in xyz]
    return torch.stack(xyz, dim=1)


@torch.jit.unused
def _handle_angle_input(
    x, dtype: torch.dtype, device: Device | None, name: str
) -> Tensor:
    """
    Helper function for building a rotation function using angles.
    The output is always of shape (N,).

    The input can be one of:
        - Torch tensor of shape (N,)
        - Python scalar
        - Torch scalar
    """
    device_ = get_device(x, device)
    if torch.is_tensor(x) and x.dim() > 1:
        msg = "Expected tensor of shape (N,); got %r (in %s)"
        raise ValueError(msg % (x.shape, name))
    return _handle_coord(x, dtype, device_)


def _assert_valid_rotation_matrix(R, tol: float = 1e-7) -> None:
    """
    Determine if R is a valid rotation matrix by checking it satisfies the
    following conditions:

    ``RR^T = I and det(R) = 1``
    """
    if R.ndim == 2:
        R = R[None, :, :]

    assert R.ndim == 3
    assert R.shape[-2:] == (3, 3)

    N = R.shape[0]

    eye = torch.eye(3, dtype=R.dtype, device=R.device)
    eye = eye.view(1, 3, 3).expand(N, -1, -1)

    orthogonal = torch.allclose(R.bmm(R.transpose(1, 2)), eye, atol=tol)
    det_R = _safe_det_3x3(R)
    no_distortion = torch.allclose(det_R, torch.ones_like(det_R))

    assert orthogonal
    assert no_distortion



from __future__ import annotations

import functools
from typing import (
    Any,
    cast,
)

import cv2
import expath
import numpy as np
import PIL.Image as pil_image
import torch
import torchvision.tv_tensors

__all__ = ["get_kwd", "read_pixels", "write_png_l16", "write_png_rgb"]

transform_context = functools.partial(
    torchvision.tv_tensors.set_return_type, "TVTensor"
)


def read_pixels(path: str, color: bool, alpha=False) -> torch.Tensor:
    """Read an image file using OpenCV."""
    if alpha:
        flags = cv2.IMREAD_UNCHANGED
    else:
        flags = (
            cv2.IMREAD_COLOR if color else cv2.IMREAD_GRAYSCALE
        ) | cv2.IMREAD_ANYDEPTH

    image = cv2.imread(path, flags)
    if image is None:
        msg = f"Failed to read image at path: {path}"
        raise RuntimeError(msg)

    if color and alpha:
        image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGBA)
    elif color:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    return torch.from_numpy(image.astype(np.int32))


def write_png_rgb(path: expath.PathType, tensor: torch.Tensor) -> None:
    """Write a tensor to a PNG file (RGB)."""
    assert tensor.ndim == 3
    if tensor.shape[0] == 3:
        tensor = tensor.permute(1, 2, 0)  # CHW -> HWC
    elif tensor.shape[-1] != 3:
        msg = "Input tensor must have 3 channels (CHW or HWC)"
        raise ValueError(msg)

    path = expath.locate(path)
    image = tensor.cpu().numpy().astype("uint8")
    pil_image.fromarray(image, mode="RGB").save(path)


def write_png_l16(path: expath.PathType, tensor: torch.Tensor) -> None:
    """Write a tensor to a PNG file (L; 16-bit)."""
    path = expath.locate(path)
    image = tensor.cpu().numpy().astype("uint16")
    pil_image.fromarray(image).save(path)


def get_kwd[KeywordType](
    kwds_dict: dict[str, Any], name: str, cast_as: type[KeywordType] | None = None, /
) -> KeywordType:
    try:
        res = kwds_dict.pop(name)
    except KeyError as e:
        msg = f"Missing keyword argument: {e.args[0]!r}"
        raise TypeError(msg) from e

    return cast(KeywordType, res)



r"""Implementation of radial spline-based linear layers."""

import itertools
import math
import typing as T

import torch
from torch import nn

from unipercept.nn.activations import ActivationSpec, get_activation
from unipercept.types import Tensor

__all__ = ["KAN", "KANLinear"]


class KANLinear(nn.Module):
    def __init__(
        self,
        in_features,
        out_features,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        standalone_scale_spline=True,
        base_activation: ActivationSpec = nn.SiLU,
        grid_eps=0.02,
        grid_range=None,
    ) -> None:
        if grid_range is None:
            grid_range = [-1, 1]
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.grid_size = grid_size
        self.spline_order = spline_order

        h = (grid_range[1] - grid_range[0]) / grid_size
        grid = (
            (
                torch.arange(-spline_order, grid_size + spline_order + 1) * h
                + grid_range[0]
            )
            .expand(in_features, -1)
            .contiguous()
        )
        self.register_buffer("grid", grid)

        self.base_weight = nn.Parameter(torch.empty(out_features, in_features))
        self.spline_weight = nn.Parameter(
            torch.empty(out_features, in_features, grid_size + spline_order)
        )
        if standalone_scale_spline:
            self.spline_scaler = nn.Parameter(torch.empty(out_features, in_features))

        self.scale_noise = scale_noise
        self.scale_base = scale_base
        self.scale_spline = scale_spline
        self.standalone_scale_spline = standalone_scale_spline
        self.activation = get_activation(base_activation)
        self.grid_eps = grid_eps

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.kaiming_uniform_(self.base_weight, a=math.sqrt(5) * self.scale_base)
        with torch.no_grad():
            noise = (
                (
                    torch.rand(self.grid_size + 1, self.in_features, self.out_features)
                    - 1 / 2
                )
                * self.scale_noise
                / self.grid_size
            )
            self.spline_weight.data.copy_(
                (self.scale_spline if not self.standalone_scale_spline else 1.0)
                * self.curve_to_coeff(
                    self.grid.T[self.spline_order : -self.spline_order],
                    noise,
                )
            )
            if self.standalone_scale_spline:
                # nn.init.constant_(self.spline_scaler, self.scale_spline)
                nn.init.kaiming_uniform_(
                    self.spline_scaler, a=math.sqrt(5) * self.scale_spline
                )

    def b_splines(self, x: Tensor) -> Tensor:
        """
        Compute the B-spline bases for the given input tensor.

        Parameters
        ----------
        x : Tnesor[B,F]
            Input tensor of shape (batch_size, in_features).

        Returns
        -------
        Tensor[B,F,G*S]
            B-spline bases tensor of shape (batch_size, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2
        assert x.size(1) == self.in_features

        grid: Tensor = self.grid  # (in_features, grid_size + 2 * spline_order + 1)
        x = x.unsqueeze(-1)
        bases = ((x >= grid[:, :-1]) & (x < grid[:, 1:])).to(x.dtype)
        for k in range(1, self.spline_order + 1):
            bases = (
                (x - grid[:, : -(k + 1)])
                / (grid[:, k:-1] - grid[:, : -(k + 1)])
                * bases[:, :, :-1]
            ) + (
                (grid[:, k + 1 :] - x)
                / (grid[:, k + 1 :] - grid[:, 1:(-k)])
                * bases[:, :, 1:]
            )

        assert bases.size() == (
            x.size(0),
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return bases.contiguous()

    def curve_to_coeff(self, x: Tensor, y: Tensor):
        """
        Compute the coefficients of the curve that interpolates the given points.

        Parameters
        ----------
        x : Tensor[B, I]
            Input tensor of shape (batch_size, in_features).
        y : Tensor[B, I, O]
            Output tensor of shape (batch_size, in_features, out_features).

        Returns
        -------
        Tensor[O,I,G*S]
            Coefficients tensor of shape (out_features, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2
        assert x.size(1) == self.in_features
        assert y.size() == (x.size(0), self.in_features, self.out_features)

        A = self.b_splines(x).transpose(
            0, 1
        )  # (in_features, batch_size, grid_size + spline_order)
        B = y.transpose(0, 1)  # (in_features, batch_size, out_features)
        solution = torch.linalg.lstsq(
            A, B
        ).solution  # (in_features, grid_size + spline_order, out_features)
        result = solution.permute(
            2, 0, 1
        )  # (out_features, in_features, grid_size + spline_order)

        assert result.size() == (
            self.out_features,
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return result.contiguous()

    @property
    def scaled_spline_weight(self):
        return self.spline_weight * (
            self.spline_scaler.unsqueeze(-1) if self.standalone_scale_spline else 1.0
        )

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        assert x.dim() == 2
        assert x.size(1) == self.in_features

        base_output = nn.functional.linear(self.activation(x), self.base_weight)
        spline_output = nn.functional.linear(
            self.b_splines(x).view(x.size(0), -1),
            self.scaled_spline_weight.view(self.out_features, -1),
        )
        return base_output + spline_output

    @torch.no_grad()
    def update_grid(self, x: Tensor, margin=0.01) -> None:
        assert x.dim() == 2
        assert x.size(1) == self.in_features
        batch = x.size(0)

        splines = self.b_splines(x)  # (batch, in, coeff)
        splines = splines.permute(1, 0, 2)  # (in, batch, coeff)
        orig_coeff = self.scaled_spline_weight  # (out, in, coeff)
        orig_coeff = orig_coeff.permute(1, 2, 0)  # (in, coeff, out)
        unreduced_spline_output = torch.bmm(splines, orig_coeff)  # (in, batch, out)
        unreduced_spline_output = unreduced_spline_output.permute(
            1, 0, 2
        )  # (batch, in, out)

        # sort each channel individually to collect data distribution
        x_sorted = torch.sort(x, dim=0)[0]
        grid_adaptive = x_sorted[
            torch.linspace(
                0, batch - 1, self.grid_size + 1, dtype=torch.int64, device=x.device
            )
        ]

        uniform_step = (x_sorted[-1] - x_sorted[0] + 2 * margin) / self.grid_size
        grid_uniform = (
            torch.arange(
                self.grid_size + 1, dtype=torch.float32, device=x.device
            ).unsqueeze(1)
            * uniform_step
            + x_sorted[0]
            - margin
        )

        grid = self.grid_eps * grid_uniform + (1 - self.grid_eps) * grid_adaptive
        grid = torch.concatenate(
            [
                grid[:1]
                - uniform_step
                * torch.arange(self.spline_order, 0, -1, device=x.device).unsqueeze(1),
                grid,
                grid[-1:]
                + uniform_step
                * torch.arange(1, self.spline_order + 1, device=x.device).unsqueeze(1),
            ],
            dim=0,
        )

        self.grid.copy_(grid.T)
        self.spline_weight.data.copy_(self.curve_to_coeff(x, unreduced_spline_output))

    def regularization_loss(self, regularize_activation=1.0, regularize_entropy=1.0):
        l1_fake = self.spline_weight.abs().mean(-1)
        regularization_loss_activation = l1_fake.sum()
        p = l1_fake / regularization_loss_activation
        regularization_loss_entropy = -torch.sum(p * p.log())
        return (
            regularize_activation * regularization_loss_activation
            + regularize_entropy * regularization_loss_entropy
        )

    if T.TYPE_CHECKING:
        __call__ = forward


class KAN(nn.Module):
    def __init__(
        self,
        layers,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        base_activation: ActivationSpec = nn.SiLU,
        grid_eps=0.02,
        grid_range=None,
    ) -> None:
        if grid_range is None:
            grid_range = [-1, 1]
        super().__init__()
        self.grid_size = grid_size
        self.spline_order = spline_order

        self.layers = nn.ModuleList(
            [
                KANLinear(
                    in_features,
                    out_features,
                    grid_size=grid_size,
                    spline_order=spline_order,
                    scale_noise=scale_noise,
                    scale_base=scale_base,
                    scale_spline=scale_spline,
                    base_activation=base_activation,
                    grid_eps=grid_eps,
                    grid_range=grid_range,
                )
                for in_features, out_features in itertools.pairwise(layers)
            ]
        )

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        for layer in self.layers:
            x = layer(x)
        return x

    if T.TYPE_CHECKING:
        __call__ = forward



import torch
from torch import nn


class MOELinear(nn.Module):
    r"""Mixture-of-experts linear layer."""

    def __init__(
        self, experts: list[nn.Module], gate: nn.Module, topk_experts: int
    ) -> None:
        super().__init__()
        assert len(experts) > 0
        self.experts = nn.ModuleList(experts)
        self.gate = gate
        self.num_experts = len(self.experts)
        self.topk_experts = topk_experts

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        gate_logits = self.gate(inputs)
        weights, selected_experts = torch.topk(gate_logits, self.topk_experts)
        weights = nn.functional.softmax(weights, dim=1, dtype=torch.float).to(
            inputs.dtype
        )
        results = torch.zeros_like(inputs)
        for i, expert in enumerate(self.experts):
            batch_idx, nth_expert = torch.where(selected_experts == i)
            results[batch_idx] += weights[batch_idx, nth_expert, None] * expert(
                inputs[batch_idx]
            )
        return results



# from __future__ import annotations
import typing as T
from typing import override

import torch
from torch import Tensor, nn
from torchvision.ops import sigmoid_focal_loss


class SigmoidFocalLoss(nn.Module):
    alpha: T.Final[float]
    gamma: T.Final[float]

    def __init__(self, *, alpha: float, gamma: float, **kwargs) -> None:
        super().__init__(**kwargs)

        self.alpha = alpha
        self.gamma = gamma

    @override
    def forward(self, x: Tensor, y: Tensor, mask: torch.Tensor | None = None) -> Tensor:
        return sigmoid_focal_loss(
            x, y, alpha=self.alpha, gamma=self.gamma, reduction="none"
        )



"""Normalization layers."""

import functools
import numbers
import typing

import torch
from torch import Tensor, nn
from torch.nn.modules.batchnorm import BatchNorm2d, SyncBatchNorm
from torchvision.ops.misc import FrozenBatchNorm2d as _FrozenBatchNorm2d

from unipercept.types import Device, DType, Size
from unipercept.utils.inspect import locate_object

# -------------------- #
# Register norm layers #
# -------------------- #

type NormFactory = typing.Callable[[int], nn.Module]
type NormSpec = str | NormFactory | type[nn.Module] | nn.Module | None


def get_default_bias(  # noqa: PLR0911
    value: bool | None, norm: nn.Module | None, *, default: bool = True
) -> bool:
    """Check whether a norm module makes the bias a transforming module obsolete.

    A bias parameter is only needed of the norm module has been explicitly set to
    not include a bias term.
    """
    # If the value is a bool, return that value
    if isinstance(value, bool):
        return value
    if norm is None:
        return default

    assert isinstance(norm, nn.Module), f"{norm=} ({type(norm)}) is not a module"

    if hasattr(norm, "default_bias"):
        return typing.cast(bool, norm.default_bias)
    if hasattr(norm, "affine"):
        return not norm.affine
    if hasattr(norm, "bias"):
        return not norm.bias
    if hasattr(norm, "elementwise_affine"):
        return not norm.elementwise_affine

    return default


def get_norm(spec: NormSpec, num_channels: int, **kwargs) -> nn.Module:
    """Resolve a norm module from a string, a factory function or a module instance.

    Parameters
    ----------
    norm
        A string, a factory function or an instance of a norm module.
    num_channels
        Number of channels to normalize.
    **kwargs
        Additional keyword arguments to pass to the norm factory.

    Notes
    -----
    When a module instance is provided,
    the ``num_channels`` argument is ignored and the object is returned as-is.
    """
    if spec is None:
        spec = nn.Identity()
    elif isinstance(spec, str):
        spec = locate_object(spec)

    # If already a module instance, return that instance directly
    if isinstance(spec, nn.Module):
        return spec
    if callable(spec):
        return spec(num_channels, **kwargs)
    msg = f"Norm module {spec!r} not found."
    raise ValueError(msg)


# --------------------- #
# LayerNorm and RMSNorm #
# --------------------- #

try:
    from apex.normalization.fused_layer_norm import (  # noqa: I001 # type: ignore[import]
        fused_layer_norm_affine,
        fused_layer_norm,
        fused_rms_norm_affine,
        fused_rms_norm,
    )

    has_apex = True
except ImportError:
    has_apex = False

LAYERNORM_DEFAULT_EPS = 1e-5
LAYERNORM_DEFAULT_MEMORY_EFFICIENT = False


def layer_norm(
    input: Tensor,
    shape: Size,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    eps: float = LAYERNORM_DEFAULT_EPS,
    memory_efficient: bool = LAYERNORM_DEFAULT_MEMORY_EFFICIENT,
) -> Tensor:
    """Layer Normalization."""
    if (
        torch.jit.is_tracing()
        or torch.jit.is_scripting()
        or torch.compiler.is_compiling()
        or not input.is_cuda
        or not has_apex
    ):
        if torch.is_autocast_enabled():
            dtype = torch.get_autocast_dtype(input.device.type)
            input = input.to(dtype)
            if weight is not None:
                weight = weight.to(dtype)
            if bias is not None:
                bias = bias.to(dtype)
        with torch.autocast(input.device.type, enabled=False):
            return nn.functional.layer_norm(input, shape, weight, bias, eps)
    if weight is not None:
        assert bias is not None
        return fused_layer_norm_affine(
            input,
            weight,
            bias,
            shape,
            eps,
            memory_efficient,
        )
    assert bias is None
    return fused_layer_norm(input, shape, eps, memory_efficient)


class LayerNorm(nn.Module):
    """Layer Normalization."""

    shape: typing.Final[tuple[int, ...]]
    eps: typing.Final[float]
    elementwise_affine: typing.Final[bool]
    memory_efficient: typing.Final[bool]

    def __init__(
        self,
        shape: typing.Iterable[int] | numbers.Integral | int,
        *,
        eps: float = LAYERNORM_DEFAULT_EPS,
        elementwise_affine: bool = True,
        device: Device = None,
        dtype: DType | None = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        if isinstance(shape, numbers.Integral | int):
            self.shape = typing.cast(tuple[int], (shape,))
        else:
            self.shape = tuple(shape)
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        self.memory_efficient = False

        if self.elementwise_affine:
            self.weight = nn.Parameter(torch.empty(self.shape, **factory_kwargs))
            self.bias = nn.Parameter(torch.empty(self.shape, **factory_kwargs))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            nn.init.zeros_(self.bias)

    @typing.override
    def forward(self, input: Tensor) -> Tensor:
        """Forward pass."""
        return layer_norm(
            input, self.shape, self.weight, self.bias, self.eps, self.memory_efficient
        )

    @typing.override
    def extra_repr(self) -> str:
        return "{shape}, eps={eps}, elementwise_affine={elementwise_affine}".format(
            **self.__dict__
        )


RMSNORM_DEFAULT_EPS = LAYERNORM_DEFAULT_EPS
RMSNORM_DEFAULT_MEMORY_EFFICIENT = LAYERNORM_DEFAULT_MEMORY_EFFICIENT


def rms_norm(
    input: Tensor,
    shape: Size,
    weight: Tensor | None = None,
    eps: float = RMSNORM_DEFAULT_EPS,
    memory_efficient: bool = RMSNORM_DEFAULT_MEMORY_EFFICIENT,
) -> Tensor:
    """Root Mean Square (RMS) Normalization."""
    if (
        torch.jit.is_tracing()
        or torch.jit.is_scripting()
        or torch.compiler.is_compiling()
        or not input.is_cuda
        or not has_apex
    ):
        if torch.is_autocast_enabled():
            dtype = torch.get_autocast_dtype(input.device.type)
            input = input.to(dtype)
            if weight is not None:
                weight = weight.to(dtype)
        with torch.autocast(input.device.type, enabled=False):
            return nn.functional.rms_norm(input, shape, weight, eps)
    if weight is not None:
        return fused_rms_norm_affine(
            input,
            weight,
            shape,
            eps,
            memory_efficient,
        )
    return fused_rms_norm(input, shape, eps, memory_efficient)


class RMSNorm(nn.Module):
    """Root Mean Square (RMS) Normalization."""

    shape: typing.Final[tuple[int, ...]]
    eps: typing.Final[float]
    elementwise_affine: typing.Final[bool]
    memory_efficient: typing.Final[bool]

    def __init__(
        self,
        shape: typing.Iterable[int] | numbers.Integral | int,
        *,
        eps: float = RMSNORM_DEFAULT_EPS,
        elementwise_affine: bool = True,
        device: Device = None,
        dtype: DType | None = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        if isinstance(shape, numbers.Integral | int):
            self.shape = typing.cast(tuple[int], (shape,))
        else:
            self.shape = tuple(shape)
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        self.memory_efficient = False
        if self.elementwise_affine:
            self.weight = nn.Parameter(torch.empty(self.shape, **factory_kwargs))
        else:
            self.register_parameter("weight", None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Reset the weights."""
        if self.elementwise_affine:
            nn.init.ones_(self.weight)

    @typing.override
    def forward(self, input: Tensor) -> Tensor:
        """Forward pass."""
        return rms_norm(input, self.shape, self.weight, self.eps, self.memory_efficient)

    @typing.override
    def extra_repr(self) -> str:
        """Extra information about the module."""
        return "{shape}, eps={eps}, elementwise_affine={elementwise_affine}".format(
            **self.__dict__
        )


# ---------- #
# Group Norm #
# ---------- #


class GroupNorm(nn.Module):
    r"""Group Normalization."""

    num_groups: typing.Final[int]
    num_channels: typing.Final[int]
    eps: typing.Final[float]
    affine: typing.Final[bool]

    def __init__(
        self,
        num_groups: int,
        num_channels: int,
        *,
        eps: float = 1e-5,
        affine: bool = True,
        device: Device = None,
        dtype: DType | None = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        if num_channels % num_groups != 0:
            msg = f"{num_channels=} must be divisible by {num_groups=}"
            raise ValueError(msg)

        self.num_groups = num_groups
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine
        if self.affine:
            self.weight = nn.Parameter(torch.empty(num_channels, **factory_kwargs))
            self.bias = nn.Parameter(torch.empty(num_channels, **factory_kwargs))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.affine:
            nn.init.ones_(self.weight)
            nn.init.zeros_(self.bias)

    @typing.override
    def forward(self, input: Tensor) -> Tensor:
        """Forward pass."""
        return nn.functional.group_norm(
            input, self.num_groups, self.weight, self.bias, self.eps
        )

    @typing.override
    def extra_repr(self) -> str:
        """Extra information about the module."""
        return "{num_groups}, {num_channels}, eps={eps}, affine={affine}".format(
            **self.__dict__
        )


def GroupNorm32(num_channels: int, **kwargs) -> nn.GroupNorm:
    """
    GroupNorm with the number of groups equal to 32, like in Detectron2.

    Notes
    -----
    The amount of channels' value (32) originates from the optimal value found in
    the GroupNorm paper.
    """
    return nn.GroupNorm(32, num_channels=num_channels, **kwargs)


def GroupNormCG(num_channels: int, **kwargs: typing.Any) -> nn.GroupNorm:
    """GroupNorm with the number of groups equal to the number of channels."""
    return nn.GroupNorm(num_channels, num_channels=num_channels, **kwargs)


def layer_norm_2d(input: Tensor, weight: Tensor, bias: Tensor, eps: float) -> Tensor:
    """Layer normalization for 2D (CHW) inputs."""
    return nn.functional.group_norm(input, 1, weight, bias, eps)


def LayerNorm2d(num_channels: int, **kwargs) -> nn.GroupNorm:
    """LayerNorm with the number of groups equal to the number of channels."""
    return nn.GroupNorm(num_groups=1, num_channels=num_channels, **kwargs)


def GroupNormFactory(
    *, num_groups: int, **kwargs
) -> typing.Callable[[int], nn.GroupNorm]:
    """GroupNorm with the number of groups equal to the number of channels."""
    return functools.partial(nn.GroupNorm, num_groups, **kwargs)


# -------------------- #
# Global Response Norm #
# -------------------- #


def global_response_norm(
    x: Tensor,
    spatial_dim: tuple[int, int],
    channel_dim: int,
    bias: Tensor,
    weight: Tensor,
    eps: float,
) -> Tensor:
    """See :class:`GlobalResponseNorm` for details."""
    x_i = x.type_as(weight)
    x_g = x_i.norm(p=2, dim=spatial_dim, keepdim=True)
    x_n = x_g / (x_g.mean(dim=channel_dim, keepdim=True) + eps)
    y = x_i + torch.addcmul(bias, weight, x_i * x_n)

    return y.type_as(x)


class GlobalResponseNorm(nn.Module):
    """Global Response Normalization based on ConvNeXt-V2.

    Paper: https://arxiv.org/abs/2301.00808
    """

    def __init__(self, channels, eps=1e-6, channels_last=True) -> None:
        super().__init__()
        self.eps = eps
        if channels_last:
            self.spatial_dim = (1, 2)
            self.channel_dim = -1
            self.wb_shape = (1, 1, 1, -1)
        else:
            self.spatial_dim = (2, 3)
            self.channel_dim = 1
            self.wb_shape = (1, -1, 1, 1)

        self.weight = nn.Parameter(torch.zeros(channels))
        self.bias = nn.Parameter(torch.zeros(channels))

    @typing.override
    def forward(self, x):
        return global_response_norm(
            x,
            self.spatial_dim,
            self.channel_dim,
            self.bias.view(self.wb_shape),
            self.weight.view(self.wb_shape),
            self.eps,
        )


class GlobalResponseNorm2d(GlobalResponseNorm):
    """Gobal response norm for CHW tensors."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, channels_last=False, **kwargs)


# ---------------- #
# Frozen BatchNorm #
# ---------------- #


class FrozenBatchNorm2d(_FrozenBatchNorm2d):
    """BatchNorm2d with frozen parameters."""

    @classmethod
    def convert_from[M: nn.Module](
        cls, module: M, recursive: bool = True
    ) -> M | typing.Self:
        """Convert BatchNorm2d or SyncBatchNorm modules to FrozenBatchNorm2d.

        See Also
        --------
        - Based on `SyncBatchNorm <https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/batchnorm.py>`_.
        """
        res = module
        if isinstance(module, BatchNorm2d | SyncBatchNorm):
            res = cls(module.num_features)
            if module.affine:
                res.weight.data = module.weight.data.clone().detach()
                res.bias.data = module.bias.data.clone().detach()
            assert module.running_mean is not None
            assert module.running_var is not None
            res.running_mean.data = module.running_mean.data
            res.running_var.data = module.running_var.data
            res.eps = module.eps
            assert module.num_batches_tracked is not None
            res.num_batches_tracked = module.num_batches_tracked
        elif recursive:
            for name, child in module.named_children():
                new_child = cls.convert_from(child, recursive=True)
                if new_child is not child:
                    res.add_module(name, new_child)
        else:
            msg = f"Cannot convert {module.__class__.__name__} to {cls.__name__}."
            raise ValueError(msg)

        del module
        return res



"""Implements wrapper functions for modifying models on the fly."""

import typing as T

import regex as re
from torch import nn

from unipercept.log import get_logger


def freeze_parameters[M: nn.Module](
    module: M, pattern: str | re.Pattern[str] | None = None, requires_grad=False
) -> M:
    """
    Freeze the parameters of a module.

    Parameters
    ----------
    module
        The module to freeze.
    pattern
        A regular expression pattern to match the parameter names that should be frozen.
        If None, all parameters are frozen.
    requires_grad
        Whether to freeze the parameters or not.
    """
    if isinstance(pattern, str):
        pattern = re.compile(pattern)
    for name, param in module.named_parameters():
        if pattern is not None and not pattern.match(name):
            continue
        get_logger(__name__).debug(f"Freezing parameter: {name}")
        param.requires_grad = requires_grad

    return module


def freeze_batchnorm[M: nn.Module](module: M, **kwargs) -> M:
    """
    Freeze the batch normalization layers of a module.

    Parameters
    ----------
    module
        The module to freeze.
    """
    from unipercept.nn.norms import FrozenBatchNorm2d

    return T.cast(M, FrozenBatchNorm2d.convert_from(module, **kwargs))


def batchnorm_to_groupnorm[M: nn.Module](module: M, num_groups: int = 32) -> M:
    """
    Convert the batch normalization layers of a module to group normalization.

    Parameters
    ----------
    module
        The module to convert.
    """
    for name, submod in module.named_modules():
        if isinstance(submod, nn.BatchNorm2d):
            num_channels = submod.num_features
            gn = nn.GroupNorm(
                min(num_groups, num_channels), num_channels, eps=submod.eps, affine=True
            )
            gn.weight.data.copy_(submod.weight.data)
            gn.bias.data.copy_(submod.bias.data)

            setattr(module, name, gn)
    return module



"""Various utility functions and classes for working with convolutional layers."""

from __future__ import annotations

import enum
import functools
import inspect
import logging
import math
import typing as T
from typing import override

import torch
import typing_extensions as TX
from torch import Tensor, nn

from unipercept.nn.activations import ActivationSpec, get_activation
from unipercept.nn.norms import NormSpec, get_norm
from unipercept.utils.function import to_2tuple

__all__ = [
    "NormActivation",
    "NormActivationMixin",
    "NormActivationModule",
    "with_norm_activation",
]


_OUTPUT_CHANNEL_PROPERTIES = [
    "out_channels",
    "out_dims",
    "channels",
    "dims",
    "d_out",
    "dim",
]


@T.overload
def get_output_channels(
    mod: nn.Module, *, use_weight: bool = True, none_ok: bool = False
) -> int: ...


@T.overload
def get_output_channels(
    mod: nn.Module, *, use_weight: bool = True, none_ok: bool = True
) -> int | None: ...


def get_output_channels(
    mod: nn.Module, *, use_weight=True, none_ok=False
) -> int | None:
    """Returns the number of output channels of a module."""
    for prop in _OUTPUT_CHANNEL_PROPERTIES:
        if not hasattr(mod, prop):
            continue
        return getattr(mod, prop)

    if use_weight and hasattr(mod, "weight"):
        return mod.weight.shape[0]

    if none_ok is True:
        return None
    msg = f"Module {mod} does not have an output channel property!"
    raise ValueError(msg)


# ---------------------- #
# Norm and activation    #
# ---------------------- #


class NormActivationWrapper(nn.Module):
    """
    Wrapper that adds normalization and activation layers to a module.

    This implementation was chosen over ``nn.Sequential`` to allow for more flexibility
    in the implementation and because this empirically improved performance in
    all tested cases that used ``torch.compile``.
    """

    def __init__(
        self, mod: nn.Module, norm: NormSpec, activation: ActivationSpec
    ) -> None:
        super().__init__()

        self.wrap = mod
        self.na = NormActivation(get_output_channels(mod), norm, activation)

    @TX.override
    def forward(self, x: Tensor) -> Tensor:
        return self.na(self.wrap(x))


class NormWrapper(nn.Module):
    """See :class:`NormActivationWrapper`."""

    def __init__(self, mod: nn.Module, norm: NormSpec) -> None:
        super().__init__()

        self.wrap = mod
        self.norm = get_norm(norm, get_output_channels(mod))

    @TX.override
    def forward(self, x: Tensor) -> Tensor:
        return self.norm(self.wrap(x))


class ActivationWrapper(nn.Module):
    """See :class:`NormActivationWrapper`."""

    def __init__(self, mod: nn.Module, activation: ActivationSpec) -> None:
        super().__init__()

        self.wrap = mod
        self.act = get_activation(activation)

    @TX.override
    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.wrap(x))


class NormActivation(nn.Module):
    def __init__(self, dim: int, norm: NormSpec, activation: ActivationSpec) -> None:
        super().__init__()

        self.norm = get_norm(norm, dim)
        self.act = get_activation(activation)

    @TX.override
    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.norm(x))


def _init_with_norm_activation(
    cls,
    *args,
    norm: NormSpec | None,
    activation: ActivationSpec | None,
    bias: bool | None = None,
    **kwargs,
) -> NormActivationWrapper:
    bias = bias if bias is not None else norm is None
    assert isinstance(bias, bool)
    mod = cls(*args, bias=bias, **kwargs)
    return NormActivationWrapper(mod, norm, activation)


def _init_with_norm(
    cls, *args, norm: NormSpec | None, bias: bool | None = None, **kwargs
) -> NormWrapper:
    bias = bias if bias is not None else norm is None
    assert isinstance(bias, bool)
    mod = cls(*args, bias=bias, **kwargs)
    return NormWrapper(mod, norm)


def _init_with_activation(
    cls, *args, activation: ActivationSpec | None = None, **kwargs
) -> ActivationWrapper:
    return ActivationWrapper(cls(*args, **kwargs), activation)


def with_norm_activation[M: nn.Module](cls: type[M]) -> type[M]:
    """Wrapper that adds normalization and activation layers to a module."""
    cls.with_activation = classmethod(_init_with_activation)  # type: ignore
    cls.with_norm = classmethod(_init_with_norm)  # type: ignore
    cls.with_norm_activation = classmethod(_init_with_norm_activation)  # type: ignore

    return cls


class NormActivationMixin:
    r"""Mixin for initialization helpers that add norm and activation layers."""

    with_activation = classmethod(_init_with_activation)
    with_norm = classmethod(_init_with_norm)
    with_norm_activation = classmethod(_init_with_norm_activation)


class NormActivationModule(nn.Module):
    r"""
    Baseclass of a module that supports normalization and activation layers.

    Equivalent to NormActivationMixin, but implemented as a base class for modules.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

    with_activation = classmethod(_init_with_activation)
    with_norm = classmethod(_init_with_norm)
    with_norm_activation = classmethod(_init_with_norm_activation)


class NormActivationSupport:
    r"""
    Mixin for initialization helpers that add norm and activation layers, but
    where the module already supports ``norm`` and ``activation`` parameters.

    The result is that the module will simply redirect the parameters to the
    instantiation of itself.
    """

    def __init__(self, *args, **kwargs) -> None:
        if "activation" in kwargs:
            msg = "Parent should handle 'activation' keyword!"
            raise NotImplementedError(msg)
        if "norm" in kwargs:
            msg = "Parent should handle 'norm' keyword!"
            raise NotImplementedError(msg)
        super().__init__(*args, **kwargs)

    @classmethod
    def with_norm(cls, *args, norm: NormSpec, **kwargs):
        if "activation" in kwargs:
            msg = "Cannot specify 'activation'!"
            raise ValueError(msg)
        return cls(*args, norm=norm, **kwargs)

    @classmethod
    def with_activation(cls, *args, activation: ActivationSpec, **kwargs):
        if "norm" in kwargs:
            msg = "Cannot specify 'norm'!"
            raise ValueError(msg)
        return cls(*args, activation=activation, **kwargs)

    @classmethod
    def with_norm_activation(
        cls, *args, norm: NormSpec, activation: ActivationSpec, **kwargs
    ):
        return cls(*args, norm=norm, activation=activation, **kwargs)


# --------------- #
# Padding support #
# --------------- #


class Padding(enum.StrEnum):
    VALID = enum.auto()
    SAME = enum.auto()


type _PaddingValue = int | tuple[int, int]
type _PaddingParam = str | Padding | _PaddingValue

_InitParams = T.ParamSpec("_InitParams")


_T = T.TypeVar("_T", bound=nn.Module)


def _wrap_init[_T: nn.Module, **InitParams](
    init: T.Callable[T.Concatenate[_T, int, int, int, _InitParams], None],
) -> T.Callable[T.Concatenate[_T, int, int, int, _InitParams], None]:
    inspect.signature(init)

    # Validate signature has first four expected arguments
    # sig_expected = ("self", "in_channels", "out_channels", "kernel_size")
    # for i, (par_expected, par) in enumerate(zip(sig_expected, sig.parameters.values())):
    #     if par_expected != par.name:
    #         raise ValueError(f"Expected parameter {par_expected} at position {i}, got {par.name} instead!")
    #     if par.kind == inspect.Parameter.KEYWORD_ONLY or par.kind == inspect.Parameter.VAR_KEYWORD:
    #         raise ValueError(f"Parameter {par.name} must be a positional argument!")

    # Update signature with new padding parameters
    # par_pad = sig.parameters.get("padding")
    # assert par_pad is not None, "Convolution module must have a padding parameter!"
    # par_pad.replace(annotation=_PaddingParam, default=Padding.SAME)

    @functools.wraps(init)
    def wrapper(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        *args: _InitParams.args,
        **kwargs: _InitParams.kwargs,
    ) -> None:
        padding: _PaddingParam = kwargs.pop("padding", Padding.SAME)  # type: ignore
        padding, dynamic = _parse_padding(padding, kernel_size, **kwargs)
        assert isinstance(padding, (int, tuple)), (
            f"Padding must be an int or a tuple, got {padding}!"
        )

        self.padding_dynamic = dynamic
        init(
            self,
            in_channels,
            out_channels,
            kernel_size,
            *args,
            padding=padding,
            **kwargs,
        )  # type: ignore

    # wrapper.__signature__ = sig  # type: ignore
    return wrapper  # type: ignore


def _wrap_forward[_T: nn.Module](
    forward: T.Callable[[_T, Tensor], Tensor],
) -> T.Callable[[_T, Tensor], Tensor]:
    @functools.wraps(forward)
    def wrapper(self, x: Tensor) -> Tensor:
        if self.padding_dynamic:
            x = _pad_same(x, self.kernel_size, self.stride, self.dilation)
        return forward(self, x)

    return wrapper


def _pad_same_amount(x: int, kernel_size: int, stride: int, dilation: int) -> int:
    """Returns the equivalent to `padding="SAME"` in TensorFlow, which supports strides greater than one."""
    # return torch.clamp(((x / stride).ceil() - 1) * stride + (kernel_size - 1) * dilation + 1 - x, min=0)
    return max(
        (math.ceil(x / stride) - 1) * stride + (kernel_size - 1) * dilation + 1 - x, 0
    )


def _pad_same(
    input: Tensor,
    kernel_size: tuple[int, ...],
    stride: tuple[int, ...],
    dilation: tuple[int, ...] = (1, 1),
    value: float = 0.0,
) -> Tensor:
    ih, iw = input.size()[-2:]
    k0, k1 = kernel_size
    s0, s1 = stride
    d0, d1 = dilation

    pad_h = _pad_same_amount(ih, k0, s0, d0)
    pad_w = _pad_same_amount(iw, k1, s1, d1)

    return nn.functional.pad(
        input,
        (pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2),
        value=value,
    )


def _parse_padding(
    padding: _PaddingParam, kernel_size, **kwargs
) -> tuple[_PaddingValue, bool]:
    """ "
    Parse the padding parameter into a value, returns -1 if the padding is not
    able to be statically determined. The second return value is a boolean that
    indicates if the padding is dynamic.
    """

    def __is_padding_static(
        kernel_size: int, stride: int = 1, dilation: int = 1, **_
    ) -> bool:
        return stride == 1 and (dilation * (kernel_size - 1)) % 2 == 0

    def __get_padding_static(
        kernel_size: int, stride: int = 1, dilation: int = 1, **_
    ) -> int:
        return ((stride - 1) + dilation * (kernel_size - 1)) // 2

    if isinstance(padding, (int, tuple)):
        return padding, False

    if isinstance(padding, str):
        padding = Padding(padding.lower())

    if isinstance(padding, Padding):
        if padding == Padding.SAME:
            if __is_padding_static(kernel_size, **kwargs):
                return __get_padding_static(kernel_size, **kwargs), False
            return 0, True
        if padding == Padding.VALID:
            return 0, False
        msg = f"Padding type {padding} not implemented yet!"
        raise NotImplementedError(msg)

    if isinstance(padding, T.Iterator) or inspect.isgenerator(padding):
        return tuple(padding), False

    msg = "Unknown padding type %s"
    raise ValueError(msg, type(padding))


def with_padding_support[_T: nn.Module](cls: type[_T]) -> type[_T]:
    cls.__init__ = _wrap_init(cls.__init__)  # type: ignore
    cls.forward = _wrap_forward(cls.forward)  # type: ignore

    return cls  # type: ignore


class PaddingMixin:
    __constants__ = "padding_dynamic"

    @override
    def __init_subclass__(cls, **kwargs) -> None:
        logging.debug(f"Initializing {cls} with padding support...")
        cls.__init__ = _wrap_init(cls.__init__)  # type: ignore
        super().__init_subclass__(**kwargs)

    @torch.jit.unused
    def _parse_padding(
        self, padding: _PaddingParam, kernel_size, **kwargs
    ) -> _PaddingValue:
        """Use during initialization to parse a padding parameter into a value."""
        parsed, self.padding_dynamic = _parse_padding(padding, kernel_size, **kwargs)

        return parsed

    def _padding_forward(
        self,
        x: Tensor,
        ks: tuple[int, ...],
        ss: tuple[int, ...],
        dl: tuple[int, ...],
    ):
        if self.padding_dynamic:
            x = _pad_same(x, ks, ss, dl)
        return x


# -------------------- #
# Other helper methods #
# -------------------- #

type _Maybe2Int = int | tuple[int, int]

# -------------------------------- #
# Pooling for convolutional layers #
# -------------------------------- #


def avg_pool2d_same(
    x,
    kernel_size: tuple[int, int],
    stride: tuple[int, int],
    ceil_mode: bool = False,
    count_include_pad: bool = True,
):
    # FIXME how to deal with count_include_pad vs not for external padding?
    x = _pad_same(x, kernel_size, stride)
    return nn.functional.avg_pool2d(
        x, kernel_size, stride, (0, 0), ceil_mode, count_include_pad
    )


class AvgPool2dSame(nn.AvgPool2d):
    """Tensorflow like 'SAME' wrapper for 2D average pooling."""

    def __init__(
        self,
        kernel_size: _Maybe2Int,
        stride: _Maybe2Int | None = None,
        ceil_mode=False,
        count_include_pad=True,
    ) -> None:
        kernel_size = to_2tuple(kernel_size)
        stride = to_2tuple(stride) if stride is not None else None
        super().__init__(kernel_size, stride, (0, 0), ceil_mode, count_include_pad)

    @override
    def forward(self, x):
        return avg_pool2d_same(
            x, self.kernel_size, self.stride, self.ceil_mode, self.count_include_pad
        )


def max_pool2d_same(
    x,
    kernel_size: tuple[int, int],
    stride: tuple[int, int],
    dilation: tuple[int, int] = (1, 1),
    ceil_mode: bool = False,
) -> Tensor:
    x = _pad_same(x, kernel_size, stride, value=-float("inf"))
    return nn.functional.max_pool2d(x, kernel_size, stride, (0, 0), dilation, ceil_mode)


class MaxPool2dSame(nn.MaxPool2d):
    """Tensorflow like 'SAME' wrapper for 2D max pooling."""

    def __init__(
        self,
        kernel_size: _Maybe2Int,
        stride: _Maybe2Int | None = None,
        dilation=1,
        ceil_mode=False,
    ) -> None:
        kernel_size = to_2tuple(kernel_size)
        stride = to_2tuple(stride) if stride is not None else None
        dilation = to_2tuple(dilation)
        super().__init__(kernel_size, stride, (0, 0), dilation, ceil_mode)

    @T.override
    def forward(self, input: Tensor) -> Tensor:
        return max_pool2d_same(
            input, self.kernel_size, self.stride, self.dilation, self.ceil_mode
        )


class MaxPool2d(nn.Module):
    def __new__(
        cls, kernel_size: _Maybe2Int, stride=None, **kwargs
    ) -> nn.MaxPool2d | MaxPool2dSame:
        stride = stride or kernel_size
        padding = kwargs.pop("padding", "")
        padding, is_dynamic = _parse_padding(
            padding, kernel_size, stride=stride, **kwargs
        )

        if is_dynamic:
            return MaxPool2dSame(kernel_size, stride=stride, **kwargs)
        return nn.MaxPool2d(kernel_size, stride=stride, padding=padding, **kwargs)


class AvgPool2d(nn.Module):
    def __new__(
        cls, kernel_size: _Maybe2Int, stride=None, **kwargs
    ) -> nn.AvgPool2d | AvgPool2dSame:
        stride = stride or kernel_size
        padding = kwargs.pop("padding", "")
        padding, is_dynamic = _parse_padding(
            padding, kernel_size, stride=stride, **kwargs
        )

        if is_dynamic:
            return AvgPool2dSame(kernel_size, stride=stride, **kwargs)
        return nn.AvgPool2d(kernel_size, stride=stride, padding=padding, **kwargs)


POOLING_LAYERS = {
    "max": MaxPool2d,
    "avg": AvgPool2d,
}



"""Generic attention and variants."""

import typing as T

import torch
from torch import Tensor, nn
from torch.nn.functional import scaled_dot_product_attention

type LinearProject = nn.Module | T.Callable[[int, int], nn.Module]


def repeat_kv(
    keys: torch.Tensor, values: torch.Tensor, groups: int, dim: int
) -> tuple[torch.Tensor, torch.Tensor]:
    r"""Used in Grouped Query Attention (GQA)."""
    keys = torch.repeat_interleave(keys, repeats=groups, dim=dim)
    values = torch.repeat_interleave(values, repeats=groups, dim=dim)
    return keys, values


class SelfAttention(nn.Module):
    def __init__(
        self,
        dim_model: int,
        dim_heads: int | None = None,
        *,
        num_heads: int = 8,
        num_groups: int = 2,
        num_sinks: int = 0,
        gate: LinearProject | None = None,
        proj: LinearProject = nn.Linear,
    ) -> None:
        r"""The most basic kind of MHA."""
        super().__init__()

        if dim_heads is None:
            assert dim_model % num_heads == 0, (dim_model, num_heads)
            dim_heads = dim_model // num_heads

        assert num_heads > 0
        assert num_groups > 0

        hidden_dim = dim_heads * num_heads

        assert hidden_dim % num_groups == 0, (hidden_dim, num_groups)

        self.num_heads = num_heads
        self.num_groups = num_groups

        self.dim_heads = dim_heads
        self.dim_model = dim_model

        self.proj_q = proj(dim_model, hidden_dim, bias=False)
        self.proj_kv = proj(dim_model, 2 * hidden_dim // num_groups, bias=False)
        self.proj_o = proj(hidden_dim, dim_model, bias=False)

        # Gate
        if gate is not None:
            self.proj_g = proj(dim_model, hidden_dim)
        else:
            self.register_module("proj_g", None)

        # Sinks
        assert num_sinks >= 0
        self.num_sinks = num_sinks
        if num_sinks > 0:
            self.sinks = nn.Parameter(
                torch.randn(num_heads // dim_heads, num_sinks, dim_heads)
                / dim_heads**0.5
            )
        else:
            self.register_parameter("sinks", None)

    def shape_inputs(self, x: Tensor, num_heads: int, head_dim: int) -> Tensor:
        batch_size = x.size(0)
        return x.reshape(batch_size, -1, num_heads, head_dim).permute(0, 2, 1, 3)

    @T.override
    def forward(self, inputs: Tensor) -> Tensor:
        r"""
        Parameters
        ----------
        x
            Query, keys and values of shape (B, N, C)

        Returns
        -------
        Tensor
            Output tensor of shape (B, N, C)
        """
        qrys = self.proj_q(inputs)
        qrys = self.shape_inputs(qrys, self.num_heads, self.dim_heads)

        keys, vals = self.proj_kv(inputs).chunk(2, dim=-1)
        keys, vals = (
            self.shape_inputs(
                x,
                self.num_heads // self.num_groups,
                self.dim_heads // self.num_groups,
            )
            for x in (keys, vals)
        )

        if self.sinks is not None:
            sinks = self.sinks.unsqueeze(0).expand(B, -1, -1, -1)
            qrys, keys, vals = (
                torch.cat((sinks, x), dim=-2) for x in (qrys, keys, vals)
            )

        keys, vals = repeat_kv(keys, vals, self.num_groups, dim=1)

        outs = scaled_dot_product_attention(
            qrys, keys, vals, enable_gqa=self.num_groups > 1
        )  # (N, H, L, E)
        outs = outs.reshape(B, -1, self.dim_heads * self.num_heads)

        if self.proj_g is not None:
            outs = self.proj_g(inputs) * outs

        return self.proj_o(outs)



r"""Camera tensors.

By convention, all spatial shapes are (H, W) and coordinates are (X, Y) where X is the
wide axis and Y is the tall axis.
"""

import functools
import typing as T

import torch
import typing_extensions as TX
from torchvision.transforms.v2.functional import register_kernel
from torchvision.transforms.v2.functional._geometry import _compute_resized_output_size
from torchvision.tv_tensors import TVTensor

from unipercept.types import Size, Tensor
from unipercept.vision.geometry import (
    AxesConvention,
    GridMode,
    apply_points,
    euclidean_to_homogeneous_points,
    extrinsics_from_parameters,
    generate_coord_grid,
    generate_fovmap,
    homogeneous_to_euclidean_points,
    intrinsics_from_parameters,
    unsafe_inverse,
)

__all__ = [
    "PinholeCameraTensor",
]


#####################
# Utility functions #
#####################


def _flip_spatial(tensor: Tensor) -> Tensor:
    r"""Flips the final (spatial) dimensions of a tensor without copying the data."""
    # assert_shape(tensor, (..., 2))
    return tensor[..., (1, 0)]


def _as_tensor[**P](wrap: T.Callable[P, Tensor]) -> T.Callable[P, Tensor]:
    r"""Converts the input to a tensor."""

    @functools.wraps(wrap)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Tensor:
        sub = wrap(*args, **kwargs)
        if torch.compiler.is_compiling():
            return sub
        return sub.as_subclass(torch.Tensor)

    return wrapper  # type: ignore


####################
# Camera container #
####################

INDEX_INTRINSICS = (..., slice(0, 4), slice(0, 4))
INDEX_EXTRINSICS = (..., slice(0, 4), slice(4, 8))
INDEX_CROP_BBOX = (..., slice(0, 4), 8)
INDEX_CANVAS_SIZE = (..., (0, 1), 9)
INDEX_SENSOR_SIZE = (..., (2, 3), 9)


class PinholeCameraTensor(TVTensor):
    """
    A PinholeCamera is a TVTensor that represents a pinhole camera model.
    This tensor always has a shape of [..., 4, 9], where the last dimension
    represents a concatenation of the intrinsic and extrinsic parameters of the camera,
    plus an additional column that represents the image size:

    - [..., :4, 0:4] represents the intrinsic matrix
    - [..., :4, 4:8] represents the extrinsic matrix
    - [..., :4, 8  ] represents the image crop bounding box as (u1, v1, u2, v2) with respect to the image size
    - [..., :2, 9  ] represents the (transformed) image size as (U, V) without cropping applied
    - [..., 2:, 9  ] represents the (original) sensor size as (U, V)
    """

    ###################
    # Factory methods #
    ###################

    @classmethod
    def with_defaults_as(
        cls,
        like: Tensor,
        n_cameras: int = 1,
    ) -> T.Self:
        """
        Constructs a PinholeCamera intiialized with heuristic values based on the input
        tensor, which we assume has the same spatial dimensions as the image.
        This should usually be avoided when the user intends to use the camera for
        rendering purposes.
        """
        if like.ndim >= 5:
            msg = (
                "Default initialization is not supported for batched tensors. This is "
                "due to that there could potentially be padding and cropping operations "
                f"that would affect the camera parameters. Got image shape: {like.shape}"
            )
            raise NotImplementedError(msg)

        H, W = like.shape[-2:]
        F = torch.tensor([1.0, 1.0]) * max(H, W) * 2
        P = torch.tensor([W / 2, H / 2])

        cam = cls.from_parameters(F, P, canvas=[W, H])
        cam = cam.unsqueeze(0)
        cam = cam.expand((n_cameras, -1, -1))
        return cam.as_subclass(cls)

    @classmethod
    def from_parameters(
        cls,
        focal_length: Tensor | T.Iterable[int | float],
        principal_point: Tensor | T.Iterable[int | float],
        angles: Tensor | T.Iterable[int | float] | None = None,  # pitch, yaw, roll
        translation: Tensor | T.Iterable[int | float] | None = None,  # x, y, z
        convention: AxesConvention | str = AxesConvention.ISO8855,
        canvas: Tensor | T.Iterable[int] | None = None,
    ) -> T.Self:
        """Constructs a PinholeCamera tensor from the camera parameters."""
        focal_length = torch.as_tensor(focal_length).float()
        principal_point = torch.as_tensor(principal_point).to(focal_length)

        intrinsics = intrinsics_from_parameters(
            focal_length,
            principal_point,
            orthographic=False,  # type: ignore
        )

        if angles is None and translation is None:
            extrinsics = None
        else:
            if angles is None:
                translation = torch.as_tensor(translation).to(focal_length)
                angles = torch.zeros_like(translation)
            else:
                angles = torch.as_tensor(angles).to(focal_length)
                translation = torch.as_tensor(translation)
            extrinsics = extrinsics_from_parameters(angles, translation, convention)

        if canvas is not None:
            canvas = torch.as_tensor(canvas).to(focal_length)

        return cls.from_parts(intrinsics, extrinsics, canvas)

    @classmethod
    def from_parts(
        cls,
        intrinsics: Tensor,
        extrinsics: Tensor | None = None,
        canvas: Tensor | None = None,
    ) -> T.Self:
        """
        Constructs a PinholeCamera tensor from the intrinsic and extrinsic matrices
        and the image size.
        """
        intrinsics = torch.as_tensor(intrinsics)
        assert intrinsics.ndim >= 2
        assert intrinsics.size(-1) == 4

        if extrinsics is None:
            extrinsics = torch.zeros_like(intrinsics)
            extrinsics[..., :, :] = torch.eye(4, device=intrinsics.device)
        else:
            extrinsics = torch.as_tensor(extrinsics)
        assert extrinsics.ndim >= 2
        assert extrinsics.size(-1) == 4

        if canvas is None:
            u0, v0 = intrinsics[..., 0, 2], intrinsics[..., 1, 2]
            canvas = torch.cat([v0 * 2, u0 * 2], dim=-1)
        else:
            canvas = torch.as_tensor(canvas).to(intrinsics)
        if canvas.size(-1) == 2:
            spatial_size = _flip_spatial(canvas)  # H, W -> W, H
            canvas = torch.zeros((*intrinsics.shape[:-2], 4, 2), dtype=intrinsics.dtype)
            canvas[..., 2:, 0] = spatial_size  # canvas bbox end
            canvas[..., :2, 1] = spatial_size  # image size
            canvas[..., 2:, 1] = spatial_size  # sensor size
        elif canvas.size(-1) != 4 or canvas.ndim != intrinsics.ndim:
            msg = (
                "Canvas must be a tensor of shape [(V, U)] or "
                "[(img_left, img_top, img_right, img_bottom), "
                "(crop_left, crop_top, crop_right, crop_bottom)]. "
                f"Got: {canvas.shape}"
            )
            raise ValueError(msg)

        res = torch.zeros((*intrinsics.shape[:-2], 4, 10), dtype=intrinsics.dtype)
        res[..., :, :4] = intrinsics
        res[..., :, 4:8] = extrinsics
        res[..., :, 8:] = canvas
        return res.as_subclass(cls)

    def __new__(
        cls,
        data: T.Any,
        *,
        dtype: torch.dtype | None = None,
        device: torch.device | str | int | None = None,
        requires_grad: bool | None = None,
    ) -> T.Self:
        tensor = cls._to_tensor(
            data, dtype=dtype, device=device, requires_grad=requires_grad
        )
        assert cls.check_pinhole_camera_tensor_shape(tensor), (
            f"Expected a tensor of shape [..., 2, 4, 4], but got {tensor.shape}"
        )
        return tensor.as_subclass(cls)

    @staticmethod
    def check_pinhole_camera_tensor_shape(tensor: Tensor) -> bool:
        """
        Returns whether the input tensor has the shape of a PinholeCamera tensor.

        This is not a sufficient condition for assessing whether the tensor is a
        PinholeCamera. It only checks the shape of the tensor and whether this would
        be a valid shape for a PinholeCamera tensor.
        """
        return tensor.size(-2) == 4 and tensor.size(-1) == 10

    ########################
    # Tensor destructuring #
    ########################

    @property
    # @_as_tensor
    def I(self) -> Tensor:
        """Returns the intrinsic matrix as a (batch_size, 4, 4) tensor."""
        return self.get_intrinsics(self)

    @property
    # @_as_tensor
    def I_inv(self) -> Tensor:
        """Returns the inverse intrinsic matrix as a (batch_size, 4, 4) tensor."""
        return unsafe_inverse(self.I)

    @property
    # @_as_tensor
    def E(self) -> Tensor:
        """Returns the extrinsic matrix as a (batch_size, 4, 4) tensor."""
        return self.get_extrinsics(self)

    @property
    # @_as_tensor
    def E_inv(self) -> Tensor:
        """Returns the inverse extrinsic matrix as a (batch_size, 4, 4) tensor."""
        return unsafe_inverse(self.E)

    @property
    # @_as_tensor
    def crop_bbox(self) -> Tensor:
        """Returns the image crop bounding box as a [B, (u1, v1, u2, v2)] tensor."""
        return self.get_crop_bbox(self)

    @property
    # @_as_tensor
    def crop_size(self) -> Tensor:
        """Returns the image size as a (batch_size, (V, U)) tensor."""
        return self.get_crop_size(self)

    @property
    # @_as_tensor
    def crop_height(self) -> Tensor:
        """Returns the image height as a (batch_size, V) tensor."""
        return self.get_crop_size(self)[..., 0]

    @property
    # @_as_tensor
    def crop_width(self) -> Tensor:
        """Returns the image width as a (batch_size, U) tensor."""
        return self.get_crop_size(self)[..., 1]

    @property
    # @_as_tensor
    def crop_center(self) -> Tensor:
        """Returns the image center as a (batch_size, (U, V)) tensor."""
        bbox = self.crop_bbox
        size = self.crop_size
        return bbox[..., :2] + _flip_spatial(size) / 2

    @property
    # @_as_tensor
    def canvas_size(self) -> Tensor:
        """Returns the canvas size as a (batch_size, (V, U)) tensor."""
        return self.get_canvas_size(self)

    @property
    # @_as_tensor
    def canvas_height(self) -> Tensor:
        """Returns the canvas height as a (batch_size, V) tensor."""
        return self.get_canvas_size(self)[..., 0]

    @property
    # @_as_tensor
    def canvas_width(self) -> Tensor:
        """Returns the canvas width as a (batch_size, U) tensor."""
        return self.get_canvas_size(self)[..., 1]

    @property
    # @_as_tensor
    def sensor_size(self) -> Tensor:
        return self.get_sensor_size(self)

    @property
    # @_as_tensor
    def sensor_height(self) -> Tensor:
        return self.sensor_size[..., 0]

    @property
    # @_as_tensor
    def sensor_width(self) -> Tensor:
        return self.sensor_size[..., 1]

    ##########################
    # Sub-matrices and views #
    ##########################

    @property
    def focal_length(self) -> Tensor:
        """Returns the focal length of the camera as a [..., (fx, fy)] tensor."""
        return self.get_focal_length(self)

    @property
    def principal_point(self) -> Tensor:
        """Returns the principal point of the camera as a [..., (u0, v0)] tensor."""
        return self.get_principal_point(self)

    @property
    def translation(self) -> Tensor:
        """Returns the translation part of the extrinsic parameters as a (batch_size, 3) tensor."""
        return self.get_translation(self)

    @property
    def K(self) -> Tensor:
        """Returns the camera matrix of the camera."""
        return self.get_intrinsics(self)[..., :3, :3]

    @property
    def R(self) -> Tensor:
        """Returns the rotation part of the extrinsic parameters as a (batch_size, 3, 3) tensor."""
        return self.get_extrinsics(self)[..., :3, :3]

    @property
    def Rt(self) -> Tensor:
        """
        Rotation-translation matrix, i.e. the concatenation :math:`[R|t]`.

        Returns
        -------
        Tensor[..., 3, 4]
            Rotation-translation matrix.
        """
        return self.get_extrinsics(self)[..., :3, :]

    @property
    def P(self) -> Tensor:
        """
        Projection matrix, defined as :math:`P = IE`.

        Returns
        -------
        Tensor[..., 4, 4]
            Projection matrix.
        """
        return self.I @ self.E

    @property
    def P_inv(self) -> Tensor:
        """
        Inverse projection matrix :math:`P^{-1}`.

        Returns
        -------
        Tensor[..., 4, 4]
            Inverse projection matrix.
        """
        return unsafe_inverse(self.P)

    ##########################
    # Representation methods #
    ##########################

    def generate_fovmap(self, like: Size | Tensor) -> Tensor:
        return generate_fovmap(self.focal_length, self.principal_point, like)

    def normalize_points(self, points_2d: Tensor) -> Tensor:
        r"""
        Normalize a set of 2D points to the camera plane.

        Returns
        -------
        Tensor[B, N, 2]
            Normalized points in the camera plane.
        """
        return (points_2d - self.principal_point) / self.focal_length

    def denormalize_points(self, points_2d: Tensor) -> Tensor:
        r"""
        Denormalize a set of 2D points to the image plane.

        Returns
        -------
        Tensor[B, N, 2]
            Denormalized points in the image plane.
        """
        return points_2d * self.focal_length + self.principal_point

    def reproject_map(self, depth_map: Tensor, noise: bool = False) -> Tensor:
        r"""
        Reprojects each pixel to 3D coordinates.

        Parameters
        ----------
        depth_map: Tensor[..., H, W]
            Depth map containing depth values for every pixel.
        noise:
            Whether to add noise to the pixel coordinates, see :func:`unipercept.vision.geometry.generate_coord_grid`.

        Returns
        -------
        Tensor[B, H, W, 3]
            Projected coordinates (XYZ) for each pixel in the depth map.
        """
        assert depth_map.ndim >= 3
        depth_map = depth_map.as_subclass(torch.Tensor)

        *batch_shape, height, width = depth_map.shape

        # Create pixel grid
        point_coords = generate_coord_grid(
            (height, width),
            device=depth_map.device,
            dtype=depth_map.dtype,
            mode=GridMode.PIXEL_NOISE if noise else GridMode.PIXEL_CENTER,
        )  # (H, W, 2)
        point_coords = point_coords.flatten(0, 1)  # (H*W, 2)

        # Reshape depth map to a depth value for every point
        point_depths = depth_map.flatten(0, -3)  # (B', H, W)
        point_depths = point_depths.reshape(-1, height * width, 1)  # (B', H*W, 1)

        # Repeat pixel coordinates for each batch item
        point_coords = point_coords.unsqueeze(0).expand(
            point_depths.size(0), -1, -1
        )  # (B', H*W, 2)

        result = self.image_to_world(point_coords, point_depths)  # (B', H*W, 3)
        result = result.unflatten(1, (height, width))
        return result.unflatten(0, batch_shape)

    def image_to_camera(
        self,
        coords: Tensor,
        depths: Tensor | None = None,
    ) -> Tensor:
        r"""
        Reprojects pixel coordinates to the camera plane. If depth is None, then
        the pixels are projected to the z=1 plane.
        """
        # points = self.normalize_points(coords)
        # points = nn.functional.normalize(points, dim=-1, p=2.0)
        # if depths is not None:
        #    points = points * depths
        points = euclidean_to_homogeneous_points(coords)
        return apply_points(self.I_inv, points) * depths

    def camera_to_world(
        self,
        points: Tensor,
    ) -> Tensor:
        r"""
        Reprojects a set of 3D points in the camera system to 3D points in the world.

        Parameters
        ----------
        points: Tensor[B, N, 3]
            2D points to reproject.

        Returns
        -------
        Tensor[B, N, 3]
            Projected coordinates (XYZ) for each pixel in the depth map.
        """
        return apply_points(self.E_inv, points)

    def image_to_world(
        self,
        coords: Tensor,
        depths: Tensor | None = None,
    ) -> Tensor:
        r"""Reprojects pixel coordinates (u,v) to 3D coordinates (x,y,z) in the world."""
        points = self.image_to_camera(coords, depths)
        return self.camera_to_world(points)

    def project_map(
        self,
        points: Tensor,
        image_shape: tuple[int, int] | Size | T.Sequence[int],
    ) -> Tensor:
        r"""
        Project a set of points onto a 2D depth map.

        Parameters
        ----------
        points: Tensor[B, N, 3]
            Batch of 3D points.
        image_shape: Tuple[int, int]
            Image shape (height, width).

        Returns
        -------
        Tensor[B, H, W]
            Depth map from the projected points.
        """
        shape = points.shape
        points = points.reshape(-1, *shape[-2:]).contiguous()

        points = apply_points(self.E, points)  # (x y z)
        points_2d = apply_points(self.I, points)  # (u v w)

        # Normalize projected points: (u v w) -> (u / w, v / w, 1)
        # points_2d = points_2d[..., :2] * points_2d[..., 2:]

        points_2d = points_2d.int()
        # Valid points inside the image plane
        # x_min, y_min, x_max, y_max = self.crop_bbox.unbind(-1)
        x_min, y_min, x_max, y_max = 0, 0, image_shape[1], image_shape[0]
        valid_mask = (
            (points_2d[..., 0] >= x_min)
            & (points_2d[..., 0] < x_max)
            & (points_2d[..., 1] >= y_min)
            & (points_2d[..., 1] < y_max)
        )

        # Calculate the flat indices of the valid pixels
        flat_points_2d = points_2d[..., 0] + points_2d[..., 1] * image_shape[1]
        flat_indices = flat_points_2d.long()

        # Create depth maps and counts using scatter_add, (B, H, W)
        depth_maps = torch.zeros([points.shape[0], *image_shape], device=points.device)
        counts = torch.zeros([points.shape[0], *image_shape], device=points.device)

        # Loop over batches to apply masks and accumulate depth/count values
        for i in range(points.shape[0]):
            valid_indices = flat_indices[i, valid_mask[i]]
            depth_maps[i].view(-1).scatter_add_(
                0, valid_indices, points[i, valid_mask[i], 2]
            )
            counts[i].view(-1).scatter_add_(
                0, valid_indices, torch.ones_like(points[i, valid_mask[i], 2])
            )

        # Calculate mean depth for each pixel in each batch
        mean_depth_maps = depth_maps / counts.clamp(min=1.0)
        return mean_depth_maps.reshape(*shape[:-2], *image_shape)

    def project_points(self, points: Tensor) -> Tensor:
        r"""
        Project a set of 3D points to 2D points.

        Parameters
        ----------
        points: Tensor[B, N, 3]
            3D points to project.

        Returns
        -------
        Tensor[B, N, 2]
            2D points projected by the camera.
        """
        result_homo = apply_points(self.P, points)
        return homogeneous_to_euclidean_points(result_homo)

    #####################
    # Modifying methods #
    #####################

    @torch.no_grad()
    def scale_(self, factor: Tensor) -> T.Self:
        """This is the inplace variant of :meth:`scale`."""
        if factor.size(-1) == 1:
            factor = factor.expand(*factor.shape[:-1], 2)
        elif factor.size(-1) != 2:
            msg = f"Expected a tensor of shape [..., (V, U)], but got {factor.shape}"
            raise ValueError(msg)

        factor_hw = factor.to(self)
        factor_wh = _flip_spatial(factor_hw)

        # Modify the intrinsic matrix
        self.principal_point[..., :] *= factor_wh
        self.focal_length[..., :] *= factor_hw

        # Modify the image size
        crop_center = self.crop_center * factor_wh
        crop_size = _flip_spatial(self.crop_size * factor_hw)

        self[INDEX_CROP_BBOX] = torch.cat(
            [crop_center - crop_size / 2, crop_center + crop_size / 2], dim=-1
        )

        # Modify the crop (and convert H, W -> V, U)
        self[INDEX_CANVAS_SIZE] *= factor_wh

        return self

    def scale(self, factor: Tensor, inplace: bool = False) -> T.Self:
        """
        Scales the camera by the given factor.

        Assumes the scaling is done w.r.t. the center of the current image crop.
        """
        if not inplace:
            self = self.clone().as_subclass(type(self))
        return self.scale_(factor)

    @torch.no_grad()
    def crop_(self, crop: Tensor) -> T.Self:
        """This is the inplace variant of :meth:`crop`."""
        crop_width = crop[..., 2] - crop[..., 0]
        crop_height = crop[..., 3] - crop[..., 1]

        # Move the principal point
        self.principal_point[..., :] -= crop[..., :2]

        assert self.canvas_width >= crop_width, (
            f"Cannot crop outside bounds: {self.canvas_width=} < {crop_width=}"
        )
        assert self.canvas_height >= crop_height, (
            f"Cannot crop outside bounds: {self.canvas_height=} < {crop_height=}"
        )

        self.crop_bbox[:] = crop.to(self)
        return self

    def crop(self, crop: Tensor, inplace: bool = False) -> T.Self:
        """
        Applies a crop to the camera.

        Parameters
        ----------
        crop: Tensor[..., (u1, v1, u2, v2)]
            The crop to apply to the camera.
        """
        if not inplace:
            self = self.clone().as_subclass(type(self))
        return self.crop_(crop)

    def horizontal_flip_(self):
        """This is the inplace variant of :meth:`horizontal_flip`."""
        self.principal_point[..., 0] = self.crop_width - self.principal_point[..., 0]

        T = torch.eye(4, device=self.device, dtype=self.dtype)
        T[..., 0, 0] = -1
        self.E[:] = self.E @ T

        return self

    def horizontal_flip(self, inplace: bool = False):
        """
        Flips the camera horizontally.

        This effectively mirrors the image along the vertical axis.
        """
        if not inplace:
            self = self.clone().as_subclass(type(self))
        return self.horizontal_flip_()

    def vertical_flip_(self):
        """This is the inplace variant of :meth:`vertical_flip`."""
        self.principal_point[..., 1] = self.crop_height - self.principal_point[..., 1]

        T = torch.eye(4, device=self.device, dtype=self.dtype)
        T[..., 1, 1] = -1
        self.E[:] = self.E @ T

        return self

    def vertial_flip(self, inplace: bool = False):
        """
        Flips the camera vertically.

        This effectively mirrors the image along the horizontal axis.
        """
        if not inplace:
            self = self.clone().as_subclass(type(self))
        return self.vertical_flip_()

    @TX.override
    def __repr__(self, *args, **kwargs) -> str:
        name = self.__class__.__name__
        if name == "Tensor":
            name = "PinholeCamera"
        kwds = ", ".join(
            f"{k}={v}"
            for k, v in {
                "shape": self.shape,
                "device": self.device,
                "requires_grad": self.requires_grad,
            }.items()
        )
        self = self.detach().cpu().as_subclass(PinholeCameraTensor)
        TAB = " " * 2
        fields = f"\n{TAB}".join(
            f"{k}: {v}"
            for k, v in {
                "focal_length": self.focal_length.tolist(),
                "principal_point": self.principal_point.tolist(),
                "rotation": self.R.tolist(),
                "translation": self.translation.tolist(),
                "canvas_size": self.canvas_size.tolist(),
                "sensor_size": self.sensor_size.tolist(),
                "crop_size": self.crop_size.tolist(),
                "crop_center": self.crop_center.tolist(),
            }.items()
        )
        return f"{self.__class__.__name__}({self.dtype}, {kwds}){{\n{TAB}{fields}\n}}"

    ##############################################################
    # Functions to get parameters, returning Tensor w/o subclass #
    ##############################################################

    @staticmethod
    @_as_tensor
    def get_intrinsics(cam: Tensor) -> Tensor:
        return cam[INDEX_INTRINSICS]

    @staticmethod
    @_as_tensor
    def get_focal_length(cam: Tensor) -> Tensor:
        return cam[..., :2, :2].diagonal(0, -2, -1)

    @staticmethod
    @_as_tensor
    def get_principal_point(cam: Tensor) -> Tensor:
        return cam[..., :2, 2]

    @staticmethod
    @_as_tensor
    def get_translation(cam: Tensor) -> Tensor:
        return cam[..., :3, 7]

    @staticmethod
    @_as_tensor
    def get_camera_matrix(cam: Tensor) -> Tensor:
        return cam[..., :3, :3]

    @staticmethod
    @_as_tensor
    def get_rotation_matrix(cam: Tensor) -> Tensor:
        return cam[..., :3, 4:7]

    @staticmethod
    @_as_tensor
    def get_extrinsics(cam: Tensor) -> Tensor:
        return cam[INDEX_EXTRINSICS]

    @staticmethod
    @_as_tensor
    def get_crop_bbox(cam: Tensor) -> Tensor:
        return cam[INDEX_CROP_BBOX]

    @staticmethod
    @_as_tensor
    def get_crop_size(cam: Tensor) -> Tensor:
        return _flip_spatial(
            PinholeCameraTensor.get_crop_bbox(cam)[..., 2:]
            - PinholeCameraTensor.get_crop_bbox(cam)[..., :2]
        )

    @staticmethod
    @_as_tensor
    def get_canvas_size(cam: Tensor) -> Tensor:
        return _flip_spatial(cam[INDEX_CANVAS_SIZE])

    @staticmethod
    @_as_tensor
    def get_sensor_size(cam: Tensor) -> Tensor:
        return _flip_spatial(cam[INDEX_SENSOR_SIZE])


#################################
# Register Torchvision handlers #
#################################


@register_kernel(functional="resize", tv_tensor_cls=PinholeCameraTensor)
def resize_camera(
    camera: PinholeCameraTensor,
    size: list[int],
    interpolation: T.Any = None,
    max_size: int | None = None,
    antialias: T.Any = False,
    use_rescale: bool = False,
) -> Tensor:
    if camera.ndim >= 3:
        shape = camera.shape[:-2]
        cams = camera.flatten(0, -3)
        resized_cams = [
            resize_camera(
                cam.as_subclass(PinholeCameraTensor),
                size,
                interpolation,
                max_size,
                antialias,
                use_rescale,
            )
            for cam in cams
        ]
        return torch.stack(resized_cams).unflatten(0, shape)

    h_old, w_old = camera.canvas_size.tolist()
    h_new, w_new = _compute_resized_output_size(
        (h_old, w_old), size=size, max_size=max_size
    )

    camera = camera.scale(torch.tensor([h_new / h_old, w_new / w_old]))
    return camera.as_subclass(PinholeCameraTensor)


@register_kernel(functional="crop", tv_tensor_cls=PinholeCameraTensor)
def crop_camera(
    camera: PinholeCameraTensor,
    top: int,
    left: int,
    height: int,
    width: int,
) -> Tensor:
    crop = torch.tensor(
        [left, top, left + width, top + height],
        device=camera.device,
        dtype=camera.dtype,
    )
    camera = camera.crop(crop)
    return camera.as_subclass(PinholeCameraTensor)


# TODO: flip
@register_kernel(functional="horizontal_flip", tv_tensor_cls=PinholeCameraTensor)
def horizontal_flip_camera(
    camera: PinholeCameraTensor,
) -> Tensor:
    return camera.horizontal_flip().as_subclass(PinholeCameraTensor)


# TODO: pad
@register_kernel(functional="vertical_flip", tv_tensor_cls=PinholeCameraTensor)
def vertical_flip_camera(
    camera: PinholeCameraTensor,
) -> Tensor:
    return camera.vertial_flip().as_subclass(PinholeCameraTensor)


PinholeCamera = PinholeCameraTensor
__all__.append("PinholeCamera")



from __future__ import annotations

from typing import override

__all__ = ["formatter"]


class formatter(property):
    """Implements a property that resolves a format string using the object's attributes."""

    __slots__ = ("fmt",)

    def __init__(self, fmt: str) -> None:
        self.fmt = fmt

    @override
    def __get__(self, obj, objtype=None):
        return self.fmt.format_map({"self": obj})

    @override
    def __set__(self, obj, value) -> None:
        msg = "Cannot set attribute"
        raise AttributeError(msg)

    @override
    def __delete__(self, obj):
        msg = "Cannot delete attribute"
        raise AttributeError(msg)



"""
Polyfill for an immutable dictionary (frozendict).

Builtin version available in Python 3.15+, after which this module can be removed.
"""

from collections.abc import Hashable, Iterable, Iterator, Mapping
from copy import deepcopy
from typing import (
    Any,
    Never,
    Self,
    TypeVar,
    cast,
    overload,
    override,
)

from unipercept.utils.descriptors import objectmagic

__all__ = ["frozendict"]

_K = TypeVar("_K", covariant=False, bound=Hashable)
_V = TypeVar("_V", covariant=True)
_X = TypeVar("_X", bound=Hashable, contravariant=True)
_Y = TypeVar("_Y", contravariant=True)


def _frozen_method(self, *args, **kwargs) -> Never:
    msg = f"'{type(self).__name__}' is frozen"
    raise TypeError(msg)


class frozendict[K: Hashable, V](dict, Mapping[K, V]):
    """
    A simple immutable dictionary.

    The API is the same as `dict`, without methods that can change the
    immutability. In addition, it supports __hash__().
    """

    __slots__ = ("_hash",)

    @objectmagic
    def __or__(
        self: Self,
        other: "frozendict[_X, _Y]",
    ) -> "frozendict[_K | _X, _V | _Y]":
        res: dict[_K | _X, _V | _Y] = {}
        res.update(dict(self.items()))
        res.update(dict(other.items()))

        return type(self)(res)

    @override
    @classmethod
    def fromkeys(cls, iterable: Iterable[_K], value: _V | None = None) -> Self:
        """See: ``dict.fromkeys()``."""
        return cls(super().fromkeys(iterable, value))

    @overload
    def __new__(cls, map_it=None, /) -> Self: ...

    @overload
    def __new__(cls, map_it: Iterable[tuple[_K, _V]] | dict[_K, _V], /) -> Self: ...

    @overload
    def __new__(cls, map_it: None, /, **kwargs: _V) -> "frozendict[str, _V]": ...

    def __new__(
        cls, map_it: Iterable[tuple[_K, _V]] | dict[_K, _V] | None = None, /, **kwargs
    ):
        if map_it is not None:
            if bool(kwargs):
                msg = "frozendict() takes no keyword arguments when a single positional argument is given."
                raise TypeError(msg)

            if map_it.__class__ == cls and cls == frozendict:
                return map_it
        else:
            map_it = cast(dict[K, V], kwargs)
        if isinstance(map_it, dict):
            it = cast(Iterable[tuple[K, V]], iter(map_it.items()))
        else:
            it = map_it

        self = super().__new__(cls, it)
        super().__init__(self, it)
        return self

    def __init__(self, *args, **kwargs) -> None:
        pass  # already initialized in __new__

    @override
    def __hash__(self, *args, **kwargs) -> int:
        try:
            return self._hash
        except AttributeError:
            pass
        try:
            h = hash(frozenset(self.items()))
        except TypeError:
            h = hash(tuple(sorted(self.items())))

        object.__setattr__(self, "_hash", h)
        return h

    @override
    def __repr__(self) -> str:
        try:
            body = super().__repr__()

            return f"frozendict({body})"
        except:
            return "frozendict"

    @override
    def __str__(self) -> str:
        return repr(self)

    @override
    def copy(self) -> Self:
        r"""Return the object itself, as it's an immutable."""
        cls = type(self)

        if cls == frozendict:
            return self

        return cls(self)

    def __copy__(self) -> Self:
        r"""See copy()."""
        return self.copy()

    def __deepcopy__(self, memo, *args, **kwargs) -> Self:
        r"""
        As for tuples, if hashable, see copy(); otherwise, it returns a
        deepcopy.
        """
        klass = type(self)
        return_copy = klass == frozendict

        if return_copy:
            try:
                hash(self)
            except TypeError:
                return_copy = False

        if return_copy:
            return self.copy()

        tmp = deepcopy(dict(self))

        return klass(tmp)

    @override
    def __reduce__(self, *args, **kwargs):
        return (type(self), (dict(self),))

    def set(self, key: _X, value: _Y) -> "frozendict[_K | _X, _V | _Y]":
        mutable_copy = dict(*self.items())
        new_self: dict[_K | _X, _Y | _V] = deepcopy(mutable_copy)
        new_self[key] = value

        return type(self)(new_self)  # type: ignore

    @override
    def setdefault(
        self: Self,
        key: _X,
        default: _Y,
    ) -> "frozendict[_K | _X, _V | _Y]":
        if key in self:
            return self  # type: ignore

        mutable_copy = dict(*self.items())
        new_self: dict[_K | _X, _Y | _V] = deepcopy(mutable_copy)
        new_self[key] = default

        return type(self)(new_self)

    def delete(self, key: Any) -> Self:
        new_self = deepcopy(dict(self))
        del new_self[key]

        if new_self:
            return type(self)(new_self)

        return type(self)()

    def _get_by_index(self, collection, index):
        try:
            return collection[index]
        except IndexError:
            maxindex = len(collection) - 1
            name = type(self).__name__
            msg = f"{name} index {index} out of range {maxindex}"
            raise IndexError(msg) from None

    def key(self, index=0) -> _K:
        collection = tuple(self.keys())

        return self._get_by_index(collection, index)

    def value(self, index=0) -> _V:
        collection = tuple(self.values())

        return self._get_by_index(collection, index)

    def item(self, index: int = 0) -> tuple[_K, _V]:
        collection = tuple(self.items())

        return self._get_by_index(collection, index)

    @override
    def __setitem__(self, key, val) -> None:
        msg = f"'{type(self).__name__}' is frozen"
        raise TypeError(msg)

    @override
    def __delitem__(self, key) -> None:
        msg = f"'{type(self).__name__}' is frozen"
        raise TypeError(msg)

    @override
    def __reversed__(self) -> Iterator[_K]:
        return reversed(tuple(self))

    clear = _frozen_method
    pop = _frozen_method
    popitem = _frozen_method
    update = _frozen_method
    __delattr__ = _frozen_method
    __setattr__ = _frozen_method



"""
This module implements functions for dealing with the state of the program.
Standard torch.distributed is used to handle distributed training and inference.
"""

from __future__ import annotations

import contextlib
import os
import sys

import torch
import torch.distributed as dist
import torch.multiprocessing as mp

__all__ = [
    "barrier",
    "check_main_process",
    "cpus_available",
    "get_device",
    "get_interactive",
    "get_process_count",
    "get_process_index",
]


###################
# Multiprocessing #
###################


def cpus_available():
    r"""
    Returns the number of CPUs available to the current process.

    This is preferred over `os.cpu_count()` as it is more reliable in HPC environments.
    """
    with contextlib.suppress(KeyError, ValueError, TypeError):
        return int(os.environ["SLURM_CPUS_PER_GPU"])

    with contextlib.suppress(Exception):
        return len(os.sched_getaffinity(0))

    return max(mp.cpu_count(), 1)


####################
# Unipercept state #
####################


def get_interactive():
    if os.getenv("UNIPERCEPT_INTERACTIVE", "").lower() in ("1", "true"):
        return True
    if sys.stdin.isatty():  # Terminal
        return True
    if hasattr(sys, "ps1"):  # IPython, Python shell, etc.
        return True
    if sys.flags.interactive:  # Python launched with -i
        return True
    return os.isatty(sys.stdout.fileno())  # Redirected output


##################################
# Data and multiprocessing utils #
##################################


def get_process_index(local: bool = False) -> int:
    """Returns the rank of the current process."""
    if not dist.is_available() or not dist.is_initialized():
        return 0
    if local:
        return int(os.environ.get("LOCAL_RANK", 0))
    return dist.get_rank()


def get_process_count() -> int:
    """Returns the total number of processes."""
    if not dist.is_available() or not dist.is_initialized():
        return 1
    return dist.get_world_size()


def check_main_process(local: bool = False) -> bool:
    """Returns True if the current process is the main process."""
    return get_process_index(local=local) == 0


def get_device() -> torch.device:
    """Returns the device assigned to the current process."""
    if torch.cuda.is_available():
        if dist.is_available() and dist.is_initialized():
            local_rank = int(os.environ.get("LOCAL_RANK", 0))
            return torch.device(f"cuda:{local_rank}")
        return torch.device("cuda:0")
    return torch.device("cpu")


def barrier():
    """Wait for all processes to reach this point."""
    if dist.is_available() and dist.is_initialized():
        dist.barrier()



"""Utilities for working with timestamps"""

import enum as E
from datetime import datetime

__all__ = ["TimestampFormat", "get_timestamp"]


class TimestampFormat(E.StrEnum):
    UNIX = E.auto()
    ISO = E.auto()
    LOCALE = E.auto()
    SHORT_YMD_HMS = E.auto()


def get_timestamp(*, format: str | TimestampFormat = TimestampFormat.ISO) -> str:
    """Returns a timestamp in the given format."""
    now = datetime.now()

    match format:
        case TimestampFormat.UNIX:
            return str(int(now.timestamp()))
        case TimestampFormat.ISO:
            return now.isoformat(timespec="seconds")
        case TimestampFormat.LOCALE:
            return now.strftime(r"%c")
        case TimestampFormat.SHORT_YMD_HMS:
            return now.strftime(r"%y%j%H%M%S")
        case _:
            msg = f"Invalid timestamp format: {format}"
            raise ValueError(msg)
